import foscat.FoCUS as FOC
import numpy as np
import foscat.Rformat as Rformat

def read(filename):
    thescat=scat(1,1,1,1)
    return thescat.read(filename)
    
class scat:
    def __init__(self,p00,s0,s1,s2,cross=False):
        self.P00=p00
        self.S0=s0
        self.S1=s1
        self.S2=s2
        self.cross=cross
        
    def get_S0(self):
        return(self.S0)

    def get_S1(self):
        return(self.S1)
    
    def get_S2(self):
        return(self.S2)

    def get_P00(self):
        return(self.P00)

    def reset_P00(self):
        self.P00=0*self.P00

    def __add__(self,other):
        assert isinstance(other, float) or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((self.P00+ other.P00), \
                        (self.S0+ other.S0), \
                        (self.S1+ other.S1), \
                        (self.S2+ other.S2))
        else:
            return scat((self.P00+ other), \
                        (self.S0+ other), \
                        (self.S1+ other), \
                        (self.S2+ other))


    def __radd__(self,other):
        return self.__add__(other)

    def __truediv__(self,other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((self.P00/ other.P00), \
                        (self.S0/ other.S0), \
                        (self.S1/ other.S1), \
                        (self.S2/ other.S2))
        else:
            return scat((self.P00/ other), \
                        (self.S0/ other), \
                        (self.S1/ other), \
                        (self.S2/ other))
        

    def __rtruediv__(self,other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((other.P00/ self.P00), \
                        (other.S0 / self.S0), \
                        (other.S1 / self.S1), \
                        (other.S2 / self.S2))
        else:
            return scat((other/ self.P00), \
                        (other / self.S0), \
                        (other / self.S1), \
                        (other / self.S2))
        
    def __sub__(self,other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((self.P00- other.P00), \
                        (self.S0- other.S0), \
                        (self.S1- other.S1), \
                        (self.S2- other.S2))
        else:
            return scat((self.P00- other), \
                        (self.S0- other), \
                        (self.S1- other), \
                        (self.S2- other))
        
    def __rsub__(self,other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((other.P00-self.P00), \
                        (other.S0- self.S0), \
                        (other.S1- self.S1), \
                        (other.S2- self.S2))
        else:
            return scat((other-self.P00), \
                        (other-self.S0), \
                        (other-self.S1), \
                        (other-self.S2))
        
    def __mul__(self,other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((self.P00* other.P00), \
                        (self.S0* other.S0), \
                        (self.S1* other.S1), \
                        (self.S2* other.S2))
        else:
            return scat((self.P00* other), \
                        (self.S0* other), \
                        (self.S1* other), \
                        (self.S2* other))

    def __rmul__(self,other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
            isinstance(other, bool) or isinstance(other, scat)
        
        if isinstance(other, scat):
            return scat((self.P00* other.P00), \
                        (self.S0* other.S0), \
                        (self.S1* other.S1), \
                        (self.S2* other.S2))
        else:
            return scat((self.P00* other), \
                        (self.S0* other), \
                        (self.S1* other), \
                        (self.S2* other))

    def plot(self,name=None,hold=True,color='blue',lw=1,legend=True):

        import matplotlib.pyplot as plt
        
        if name is None:
            name=''

        if hold:
            plt.figure(figsize=(8,8))
        
        plt.subplot(2,2,1)
        if legend:
            plt.plot(self.get_np(abs(self.S0)).flatten(),color=color,label=r'%s $S_0$'%(name),lw=lw)
        else:
            plt.plot(self.get_np(abs(self.S0)).flatten(),color=color,lw=lw)
        plt.yscale('log')
        plt.legend()
        plt.subplot(2,2,2)
        if legend:
            plt.plot(self.get_np(abs(self.S1)).flatten(),color=color,label=r'%s $S_1$'%(name),lw=lw)
        else:
            plt.plot(self.get_np(abs(self.S1)).flatten(),color=color,lw=lw)
        plt.yscale('log')
        plt.legend()
        plt.subplot(2,2,3)
        if legend:
            plt.plot(self.get_np(abs(self.P00)).flatten(),color=color,label=r'%s $P_{00}$'%(name),lw=lw)
        else:
            plt.plot(self.get_np(abs(self.P00)).flatten(),color=color,lw=lw)
        plt.yscale('log')
        plt.legend()
        plt.subplot(2,2,4)
        if legend:
            plt.plot(self.get_np(abs(self.S2)).flatten(),color=color,label=r'%s $S_2$'%(name),lw=lw)
        else:
            plt.plot(self.get_np(abs(self.S2)).flatten(),color=color,lw=lw)
        plt.yscale('log')
        plt.legend()
        
    def save(self,filename):
        np.save('%s_s0.npy'%(filename), self.get_S0().numpy())
        np.save('%s_s1.npy'%(filename), self.get_S1().numpy())
        np.save('%s_s2.npy'%(filename), self.get_S2().numpy())
        np.save('%s_p0.npy'%(filename), self.get_P00().numpy())
        
    def read(self,filename):
        s0=np.load('%s_s0.npy'%(filename))
        s1=np.load('%s_s1.npy'%(filename))
        s2=np.load('%s_s2.npy'%(filename))
        p0= np.load('%s_p0.npy'%(filename))
        return scat(p0,s0,s1,s2)
    
    def get_np(self,x):
        if isinstance(x, np.ndarray):
            return x
        else:
            return x.numpy()

    def std(self):
        return np.sqrt(((abs(self.get_np(self.S0)).std())**2+(abs(self.get_np(self.S1)).std())**2+(abs(self.get_np(self.S2)).std())**2+(abs(self.get_np(self.P00)).std())**2)/4)

    def mean(self):
        return abs(self.get_np(self.S0).mean()+self.get_np(self.S1).mean()+self.get_np(self.S2).mean()+self.get_np(self.P00).mean())/3
        
        
    
class funct(FOC.FoCUS):
    
    def eval(self, image1, image2=None,mask=None,Auto=True,s0_off=1E-6):

        ### AUTO OR CROSS
        cross = False
        if image2 is not None:
            cross = True
            all_cross=not Auto
        else:
            all_cross=False
            
        axis=1
        
        # determine jmax and nside corresponding to the input map
        im_shape = image1.shape
        if self.use_R_format and isinstance(image1,Rformat.Rformat):
            if len(image1.shape)==4:
                nside=im_shape[2]-2*self.R_off
                npix = self.chans*nside*nside # Number of pixels
            else:
                nside=im_shape[1]-2*self.R_off
                npix = self.chans*nside*nside  # Number of pixels
        else:
            if len(image1.shape)==2:
                npix = int(im_shape[1])  # Number of pixels
            else:
                npix = int(im_shape[0])  # Number of pixels

            if self.chans==1:
                nside=im_shape[axis]
                npix=nside*nside
            else:
                nside=int(np.sqrt(npix//self.chans))
                
        jmax=int(np.log(nside)/np.log(2))-self.OSTEP

        ### LOCAL VARIABLES (IMAGES and MASK)
        # Check if image1 is [Npix] or [Nbatch,Npix]
        if len(image1.shape)==1 or (len(image1.shape)==2 and self.chans==1) or (len(image1.shape)==3 and isinstance(image1,Rformat.Rformat)):
            # image1 is [Nbatch, Npix]
            I1 = self.backend.bk_cast(self.backend.bk_expand_dims(image1,0))  # Local image1 [Nbatch, Npix]
            if cross:
                I2 = self.backend.bk_cast(self.backend.bk_expand_dims(image2,0))  # Local image2 [Nbatch, Npix]
        else:
            I1=self.backend.bk_cast(image1)
            if cross:
                I2=self.backend.bk_cast(image2)
                
        # self.mask is [Nmask, Npix]
        if mask is None:
            if self.chans==1:
                vmask = self.backend.bk_ones([1,nside,nside],dtype=self.all_type)
            else:
                vmask = self.backend.bk_ones([1,npix],dtype=self.all_type)
            
            if self.use_R_format:
                vmask = self.to_R(vmask,axis=1,chans=self.chans)
        else:
            vmask = self.backend.bk_cast( mask) # [Nmask, Npix]
            
            if self.use_R_format:
                vmask=self.to_R(vmask,axis=1,chans=self.chans)

        if self.use_R_format:
            I1=self.to_R(I1,axis=axis,chans=self.chans)
            if cross:
                I2=self.to_R(I2,axis=axis,chans=self.chans)

        if self.KERNELSZ>3:
            # if the kernel size is bigger than 3 increase the binning before smoothing
            l_image1=self.up_grade(I1,nside*2,axis=axis)
            vmask=self.up_grade(vmask,nside*2,axis=1)
            if cross:
                l_image2=self.up_grade(I2,nside*2,axis=axis)
        else:
            l_image1=I1
            if cross:
                l_image2=I2

        s0=None
        s0 = self.masked_mean(l_image1,vmask,axis=axis)+s0_off
        
        if cross:
            if len(image1.shape)==1 or (len(image1.shape)==3 and isinstance(image1,Rformat.Rformat)):
                s0 = self.backend.bk_concat([s0,self.masked_mean(l_image2,vmask,axis=axis)],1)+s0_off
            else:
                s0 = self.backend.bk_concat([s0,self.masked_mean(l_image2,vmask,axis=axis)],0)+s0_off

        s1=None
        s2=None
        p00=None
        l2_image=None
        l2_image_imag=None

        for j1 in range(jmax):
            # Convol image along the axis defined by 'axis' using the wavelet defined at
            # the foscat initialisation
            #c_image_real is [....,Npix_j1,....,Norient]
            c_image1=self.convol(l_image1,axis=axis)
            if cross:
                c_image2=self.convol(l_image2,axis=axis)
            else:
                c_image2=c_image1

            # Compute (a+ib)*(a+ib)* the last c_image column is the real and imaginary part
            conj=c_image1*self.backend.bk_conjugate(c_image2)
            if Auto:
                conj=self.backend.bk_real(conj)

            # Compute l_p00 [....,....,Nmask,1,Norient]  
            l_p00 = self.backend.bk_expand_dims(self.masked_mean(conj,vmask,axis=axis),-2)

            conj=self.backend.bk_L1(conj)

            # Compute l_s1 [....,....,Nmask,1,Norient] 
            l_s1 = self.backend.bk_expand_dims(self.masked_mean(conj,vmask,axis=axis),-2)

            # Concat S1,P00 [....,....,Nmask,j1,Norient] 
            if s1 is None:
                s1=l_s1
                p00=l_p00
            else:
                s1=self.backend.bk_concat([s1,l_s1],axis=-2)
                p00=self.backend.bk_concat([p00,l_p00],axis=-2)

            # Concat l2_image [....,Npix_j1,....,j1,Norient]
            if l2_image is None:
                l2_image=self.backend.bk_expand_dims(self.update_R_border(conj,axis=axis),axis=-2)
            else:
                l2_image=self.backend.bk_concat([self.backend.bk_expand_dims(self.update_R_border(conj,axis=axis),axis=-2),l2_image],axis=-2)

            # Convol l2_image [....,Npix_j1,....,j1,Norient,Norient]
            c2_image=self.convol(self.backend.bk_relu(l2_image),axis=axis)

            conj2p=self.backend.bk_L1(c2_image*self.backend.bk_conjugate(c2_image))
            if Auto:
                conj2p=self.backend.bk_real(conj2p)

            c2_image=self.convol(self.backend.bk_relu(-l2_image),axis=axis)

            conj2m=self.backend.bk_L1(c2_image*self.backend.bk_conjugate(c2_image))
            if Auto:
                conj2m=self.backend.bk_real(conj2m)

            # Convol l_s2 [....,....,Nmask,j1,Norient,Norient]
            l_s2 = self.masked_mean(conj2p-conj2m,vmask,axis=axis)

            # Concat l_s2 [....,....,Nmask,j1*(j1+1)/2,Norient,Norient]
            if s2 is None:
                s2=l_s2
            else:
                s2=self.backend.bk_concat([s2,l_s2],axis=-3)

            if j1!=jmax-1:
                # Rescale vmask [Nmask,Npix_j1//4]   
                vmask = self.smooth(vmask,axis=1)
                vmask = self.ud_grade_2(vmask,axis=1)

                # Rescale l2_image [....,Npix_j1//4,....,j1,Norient]   
                l2_image = self.smooth(l2_image,axis=axis)
                l2_image = self.ud_grade_2(l2_image,axis=axis)

                # Rescale l_image [....,Npix_j1//4,....]  
                l_image1 = self.smooth(l_image1,axis=axis)
                l_image1 = self.ud_grade_2(l_image1,axis=axis)
                if cross:
                    l_image2 = self.smooth(l_image2,axis=axis)
                    l_image2 = self.ud_grade_2(l_image2,axis=axis)

        if len(image1.shape)==1 or (len(image1.shape)==3 and isinstance(image1,Rformat.Rformat)):
            return(scat(p00[0],s0[0],s1[0],s2[0],cross))

        return(scat(p00,s0,s1,s2,cross))

    def square(self,x):
        # the abs make the complex value usable for reduce_sum or mean
        return scat(self.backend.bk_square(self.backend.bk_abs(x.P00)),
                    self.backend.bk_square(self.backend.bk_abs(x.S0)),
                    self.backend.bk_square(self.backend.bk_abs(x.S1)),
                    self.backend.bk_square(self.backend.bk_abs(x.S2)))

    def reduce_mean(self,x,axis=None):
        if axis is None:
            return  (self.backend.bk_reduce_mean(x.P00)+
                     self.backend.bk_reduce_mean(x.S0)+
                     self.backend.bk_reduce_mean(x.S1)+
                     self.backend.bk_reduce_mean(x.S2))/4
        else:
            return (self.backend.bk_reduce_mean(x.P00,axis=axis)+
                    self.backend.bk_reduce_mean(x.S0,axis=axis)+
                    self.backend.bk_reduce_mean(x.S1,axis=axis)+
                    self.backend.bk_reduce_mean(x.S2,axis=axis))/4

    def reduce_sum(self,x,axis=None):
        if axis is None:
            return  self.backend.bk_reduce_sum(x.P00)+ \
                self.backend.bk_reduce_sum(x.S0)+ \
                self.backend.bk_reduce_sum(x.S1)+ \
                self.backend.bk_reduce_sum(x.S2)
        else:
            return scat(self.backend.bk_reduce_sum(x.P00,axis=axis),
                        self.backend.bk_reduce_sum(x.S0,axis=axis),
                        self.backend.bk_reduce_sum(x.S1,axis=axis),
                        self.backend.bk_reduce_sum(x.S2,axis=axis))
            

    def log(self,x):
        return scat(self.backend.bk_log(x.P00),
                    self.backend.bk_log(x.S0),
                    self.backend.bk_log(x.S1),
                    self.backend.bk_log(x.S2))
    def inv(self,x):
        return scat(1/(x.P00),1/(x.S0),1/(x.S1),1/(x.S2))

    def one(self):
        return scat(1.0,1.0,1.0,1.0)
        



def genDemo():	
	
	#=============================	
	
	f=open("test2Dplot.py","w")

	f.write("import numpy as np\n")
	f.write("import matplotlib.pyplot as plt\n")
	f.write("\n")
	f.write("outname='test2D'\n")
	f.write("\n")
	f.write("scratch_path = 'foscat_data'\n")
	f.write("r1=np.load(scratch_path+'/%s_r1.npy'%(outname))\n")
	f.write("r2=np.load(scratch_path+'/%s_r2.npy'%(outname))\n")
	f.write("o1=np.load(scratch_path+'/%s_o1.npy'%(outname))\n")
	f.write("o2=np.load(scratch_path+'/%s_o2.npy'%(outname))\n")
	f.write("\n")
	f.write("plt.figure(figsize=(12,8))\n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.plot(r1.flatten(),color='blue',label='input')\n")
	f.write("plt.plot(o1.flatten(),color='red',label='synthesised')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.ylabel(r'$S_1$')\n")
	f.write("plt.legend()\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.plot(r2.flatten(),color='blue',label='input')\n")
	f.write("plt.plot(o2.flatten(),color='red',label='synthesised')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.ylabel(r'$S_2$')\n")
	f.write("\n")
	f.write("ref =np.load(scratch_path+'/%s_ref.npy'%(outname))\n")
	f.write("omap=np.load(scratch_path+'/%s_result.npy'%(outname))\n")
	f.write("\n")
	f.write("amp=ref.max()\n")
	f.write("nx=ref.shape[0]\n")
	f.write("plt.figure(figsize=(12,8))\n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.imshow(ref[nx//2-nx//4:nx//2+nx//4,nx//2-nx//4:nx//2+nx//4],cmap='jet',origin='lower',vmin=-amp,vmax=amp)\n")
	f.write("plt.title('Input')\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.imshow(omap[nx//2-nx//4:nx//2+nx//4,nx//2-nx//4:nx//2+nx//4],cmap='jet',origin='lower',vmin=-amp,vmax=amp)\n")
	f.write("plt.title('FoSCAT')\n")
	f.write("plt.show()\n")
	
	
	#=============================	
	
	f=open("test2D.py","w")

	f.write("import numpy as np\n")
	f.write("import os, sys\n")
	f.write("import matplotlib.pyplot as plt\n")
	f.write("import foscat.FoCUS as FOC\n")
	f.write("\n")
	f.write("def conjugate(mat,vec,nitt=10):\n")
	f.write("    x=np.zeros([vec.shape[0]])\n")
	f.write("    r=vec-np.dot(mat,x)\n")
	f.write("    p=r\n")
	f.write("    for itt in range(nitt):\n")
	f.write("        Ap=np.dot(mat,p)\n")
	f.write("        delta=np.sum(r*r)\n")
	f.write("        if itt%10==9:\n")
	f.write("            print(itt,delta)\n")
	f.write("        alpha=delta/np.sum(p*Ap)\n")
	f.write("        x=x+alpha*p\n")
	f.write("        r=r-alpha*Ap\n")
	f.write("        beta=np.sum(r*r)/delta\n")
	f.write("        p=r+beta*p\n")
	f.write("    return(x)\n")
	f.write("\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# INITIALIZE foscat class\n")
	f.write("#=================================================================================\n")
	f.write("#  Change the temporary path\n")
	f.write("#=================================================================================\n")
	f.write("scratch_path = 'foscat_data'\n")
	f.write("outname='test2D'\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# This test could be run till XSIZE=512 but could last for Hours if not GPUs are\n")
	f.write("# available. XSIZE=128 takes few minutes with NVIDIA T4.\n")
	f.write("#=================================================================================\n")
	f.write("XSIZE=64\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("#==   Make the statistic invariant by rotation => Higher dimension reduction    ==\n")
	f.write("#==   avg_ang = True                                                            ==\n")
	f.write("#=================================================================================\n")
	f.write("avg_ang=False\n")
	f.write("\n")
	f.write("fc=FOC.FoCUS(healpix=False,NORIENT=4,KERNELSZ=3,OSTEP=0,TEMPLATE_PATH=scratch_path,slope=2)\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("#  READ data and get data if necessary: \n")
	f.write("#=================================================================================\n")
	f.write("#  Here the input data is a vorticity map of 512x512:\n")
	f.write("# This image shows a simulated snapshot of ocean turbulence in the North Atlantic Ocean in March 2012,\n")
	f.write("# from a groundbreaking super-high-resolution global ocean simulation (approximately 1.2 miles, \n")
	f.write("#or 2 kilometers, horizontal resolution) developed at JPL.\n")
	f.write("# (http://wwwcvs.mitgcm.org/viewvc/MITgcm/MITgcm_contrib/llc_hires/llc_4320/). \n")
	f.write("#=================================================================================\n")
	f.write("try:\n")
	f.write("    d=np.load(scratch_path+'/Vorticity.npy')\n")
	f.write("except:\n")
	f.write("    import imageio as iio\n")
	f.write("\n")
	f.write("    os.system('wget -O '+scratch_path+'/PIA22256.tif https://photojournal.jpl.nasa.gov/tiff/PIA22256.tif')\n")
	f.write("    \n")
	f.write("    im=iio.imread(scratch_path+'/PIA22256.tif')\n")
	f.write("    im=im[1000:1512,2000:2512,0]/255.0-im[1000:1512,2000:2512,2]/255.0\n")
	f.write("    np.save(scratch_path+'/Vorticity.npy',im)\n")
	f.write("    os.system('rm '+scratch_path+'/PIA22256.tif')\n")
	f.write("    d=np.load(scratch_path+'/Vorticity.npy')\n")
	f.write("\n")
	f.write("\n")
	f.write("d=d[0:XSIZE,0:XSIZE]\n")
	f.write("nx=d.shape[0]\n")
	f.write("# define the level of noise of the simulation\n")
	f.write("ampnoise=1.0\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Synthesise data with the same cross statistic than the input data\n")
	f.write("#=================================================================================\n")
	f.write("\n")
	f.write("# convert data in tensor for Foscat\n")
	f.write("idata = fc.convimage(d)\n")
	f.write("\n")
	f.write("# define the mask where the statistic are used\n")
	f.write("x=np.repeat((np.arange(XSIZE)-XSIZE/2)/XSIZE,XSIZE).reshape(XSIZE,XSIZE)\n")
	f.write("mask=np.exp(-32*(x**4+(x.T)**4))\n")
	f.write("mask[:,:]=1.0\n")
	f.write("#mask[32,33]=1.0\n")
	f.write("fc.add_mask(mask.reshape(1,nx,nx))\n")
	f.write("\n")
	f.write("# Initialize the learning and initialize the tensor to be synthesized\n")
	f.write("randfield=np.random.randn(nx,nx)\n")
	f.write("#randfield[:,:]=0.0\n")
	f.write("#randfield[32:34,32]=1.0\n")
	f.write("ldata=fc.init_synthese(randfield)\n")
	f.write("\n")
	f.write("# Build the loss:\n")
	f.write("# here Loss += (d x d - s x s - tb[0]).tw[0]\n")
	f.write("fc.add_loss_2d(idata,idata,ldata,ldata,avg_ang=avg_ang)\n")
	f.write("\n")
	f.write("# initiliaze the synthesise process\n")
	f.write("loss=fc.init_optim()\n")
	f.write("\n")
	f.write("tw1={}\n")
	f.write("tw2={}\n")
	f.write("tb1={}\n")
	f.write("tb2={}\n")
	f.write("# compute the weights and the bias for each loss\n")
	f.write("modd=d.reshape(1,nx,nx)\n")
	f.write("r1,r2=fc.calc_stat(modd,modd,avg_ang=avg_ang)\n")
	f.write("\n")
	f.write("tw1[0]=1.0+0.*r1[0]\n")
	f.write("tw2[0]=1.0+0.*r2[0]\n")
	f.write("tb1[0]=0.0*r1[0]\n")
	f.write("tb2[0]=0.0*r2[0]\n")
	f.write("\n")
	f.write("#save the reference statistics\n")
	f.write("np.save(scratch_path+'/%s_r1.npy'%(outname), r1)\n")
	f.write("np.save(scratch_path+'/%s_r2.npy'%(outname), r2)\n")
	f.write("feed_dict={}\n")
	f.write("feed_dict[fc.tw1[0]]=tw1[0]\n")
	f.write("feed_dict[fc.tw2[0]]=tw2[0]\n")
	f.write("feed_dict[fc.tb1[0]]=tb1[0]\n")
	f.write("feed_dict[fc.tb2[0]]=tb2[0]\n")
	f.write("\n")
	f.write("sr1,sr2=fc.cwst1_comp(idata,idata)\n")
	f.write("s1,s2=fc.cwst1_comp(ldata,ldata)\n")
	f.write("\n")
	f.write("r1,r2=fc.sess.run([sr1,sr2],feed_dict=feed_dict)\n")
	f.write("rr1n,rr2n=fc.sess.run([s1,s2],feed_dict=feed_dict)\n")
	f.write("\n")
	f.write("print('initialize grad 1')\n")
	f.write("ngrd1=5\n")
	f.write("g2={}\n")
	f.write("for i in range(4):\n")
	f.write("    for j in range(5):\n")
	f.write("        g2[i+4*j]=fc.opti.compute_gradients(s1[j][0,i],var_list=[fc.param[0]])\n")
	f.write("\n")
	f.write("print('initialize grad 2')\n")
	f.write("ngrad=14\n")
	f.write("g22={}\n")
	f.write("for i in range(16):\n")
	f.write("    for j in range(ngrad):\n")
	f.write("        g22[i+16*j]=fc.opti.compute_gradients(s2[j][0,i//4,i%4],var_list=[fc.param[0]])\n")
	f.write("print('initialize grad Done')\n")
	f.write("\n")
	f.write("xidx={}\n")
	f.write("yidx={}\n")
	f.write("for i in range(5):\n")
	f.write("    xidx[i]=(np.arange(64*64,dtype='int')//64)//(2**i)\n")
	f.write("    yidx[i]=(np.arange(64*64,dtype='int')%64)//(2**i)\n")
	f.write("   \n")
	f.write("for itt in range(10000):\n")
	f.write("    \n")
	f.write("    rr1,rr2=fc.sess.run([s1,s2],feed_dict=feed_dict)\n")
	f.write("    \n")
	f.write("    gg1=fc.sess.run(g2,feed_dict=feed_dict)\n")
	f.write("    gg2=fc.sess.run(g22,feed_dict=feed_dict)\n")
	f.write("    \n")
	f.write("    diff=np.zeros([1,64,64,1])\n")
	f.write("    \n")
	f.write("    for i in range(5):\n")
	f.write("        for j in range(4):\n")
	f.write("            tmp=(rr1[i][0,j]-r1[i][0,j])*(gg1[i*4+j][0][0].reshape(64,64)/(64*64))\n")
	f.write("            diff[0,:,:,0]+=tmp/64 #(tmp[xidx[i],yidx[i]].reshape(64,64)/64\n")
	f.write("    \n")
	f.write("    for i in range(ngrad):\n")
	f.write("        for j in range(16):\n")
	f.write("            tmp=(rr2[i][0,j//4,j%4]-r2[i][0,j//4,j%4])*(gg2[i*16+j][0][0].reshape(64,64))/(64*64)\n")
	f.write("            diff[0,:,:,0]+=tmp/2048 #(tmp[xidx[i],yidx[i]].reshape(64,64))/64\n")
	f.write("            \n")
	f.write("    par=fc.get_param()\n")
	f.write("    fc.set_value(par-(diff).flatten(),0)\n")
	f.write("    if itt%100==0:\n")
	f.write("        var1=0\n")
	f.write("        for i in range(5):\n")
	f.write("            var1+=((r1[i]-rr1[i]).std())**2\n")
	f.write("        var2=0\n")
	f.write("        for i in range(ngrad):\n")
	f.write("            var2+=((r2[i]-rr2[i]).std())**2\n")
	f.write("        print(itt,np.sqrt(var1),np.sqrt(var2))\n")
	f.write("        """\n")
	f.write("        plt.figure()\n")
	f.write("        plt.subplot(1,2,1)\n")
	f.write("        plt.plot(rr1n.flatten(),color='yellow',lw=8)\n")
	f.write("        plt.plot(r1.flatten(),color='black')\n")
	f.write("        plt.plot(rr1.flatten(),color='blue')\n")
	f.write("        plt.yscale('log')\n")
	f.write("        plt.subplot(1,2,2)\n")
	f.write("        plt.plot(rr2n.flatten(),color='yellow',lw=8)\n")
	f.write("        plt.plot(r2.flatten(),color='black')\n")
	f.write("        plt.plot(rr2.flatten(),color='blue')\n")
	f.write("        plt.yscale('log')\n")
	f.write("        plt.show()\n")
	f.write("        """\n")
	f.write("imap=fc.get_map()\n")
	f.write("plt.figure(figsize=(16,8))\n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.imshow(d.reshape(64,64),cmap='jet')\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.imshow(imap.reshape(64,64),cmap='jet')\n")
	f.write("plt.show()\n")
	f.write("exit(0)\n")
	f.write("\n")
	f.write("exit(0)\n")
	f.write("gr1=fc.sess.run(g1,feed_dict=feed_dict)\n")
	f.write("gc1=fc.sess.run(grad1,feed_dict=feed_dict)\n")
	f.write("gr2=fc.sess.run(g2,feed_dict=feed_dict)\n")
	f.write("gr22=fc.sess.run(g22,feed_dict=feed_dict)\n")
	f.write("gc2=fc.sess.run(grad2,feed_dict=feed_dict)\n")
	f.write("gc3=fc.sess.run(grad3,feed_dict=feed_dict)\n")
	f.write("\n")
	f.write("c,s=fc.get_ww()\n")
	f.write("np.save('coef.npy',c)\n")
	f.write("np.save('coef1.npy',gc1[0])\n")
	f.write("np.save('coef2.npy',gc2[0])\n")
	f.write("np.save('coef3.npy',gc3[0])\n")
	f.write("\n")
	f.write("c=c.reshape(3,3,4)\n")
	f.write("def convol(val,c):\n")
	f.write("    val2=np.zeros([val.shape[0]+2,val.shape[1]+2])\n")
	f.write("    val2[1:-1,1:-1]=val\n")
	f.write("    res=0*val\n")
	f.write("    for i in range(3):\n")
	f.write("        for j in range(3):\n")
	f.write("            res=res+c[i,j]*val2[i:-2+i,j:-2+j]\n")
	f.write("    return(res)\n")
	f.write("\n")
	f.write("#rrr=convol(np.ones([64,64]),c[:,:,0])\n")
	f.write("#print(rrr[0:3,0:3])\n")
	f.write("\n")
	f.write("print(s2[0])\n")
	f.write("print(gc2[0].shape)\n")
	f.write("\n")
	f.write("print(c.shape)\n")
	f.write("for i in range(4):\n")
	f.write("    print('=====  GRAD S1 CALC')\n")
	f.write("    print(gc1[0][0,30:35,30:35,i])\n")
	f.write("    print('=====  GRAD S1 TENS')\n")
	f.write("    print((gr2[i][0][0].reshape(64,64))[30:35,30:35])\n")
	f.write("    \n")
	f.write("for i in range(4):\n")
	f.write("    print('===== GRAD S2 CALC',i//4,i%4)\n")
	f.write("    print(gc2[0][0,30:35,30:35,i])\n")
	f.write("    print('....')\n")
	f.write("    print((gr22[i][0][0].reshape(64,64))[30:35,30:35])\n")
	f.write("\n")
	f.write("exit(0)\n")
	f.write("plt.figure(figsize=(16,6))\n")
	f.write(""""\n")
	f.write("plt.subplot(2,1,1)\n")
	f.write("plt.plot(gr1[0][0],color='blue',lw=8)\n")
	f.write("plt.plot(gc1[0][0,:,:,1].flatten(),color='red')\n")
	f.write(""""\n")
	f.write("for i in range(4):\n")
	f.write("    plt.subplot(2,4,1+i)\n")
	f.write("    plt.imshow(gr22[i][0][0].reshape(64,64),cmap='jet')\n")
	f.write("    plt.subplot(2,4,5+i)\n")
	f.write("    plt.imshow(-gc2[0][0,:,:,4*i].reshape(64,64),cmap='jet')\n")
	f.write("#plt.plot(gr2[0][0],color='blue',lw=8)\n")
	f.write("#plt.plot(gc2[0][0,:,:,0].flatten(),color='red')\n")
	f.write("plt.show()\n")
	f.write("exit(0)\n")
	f.write("\n")
	f.write("gg=fc.sess.run(grad,feed_dict=feed_dict)\n")
	f.write("g2=fc.sess.run(grad2,feed_dict=feed_dict)\n")
	f.write("\n")
	f.write("imap=fc.get_map()\n")
	f.write("imm=imap.reshape(1,64,64,1)\n")
	f.write("or1,or2=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("imm[0,10,10,0]+=1E-6\n")
	f.write("rr1,rr2=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write(""""\n")
	f.write("plt.figure()\n")
	f.write("for i in range(len(gg)):\n")
	f.write("    for k in range(4):\n")
	f.write("        plt.subplot(5,4,1+4*i+k)\n")
	f.write("        plt.imshow(gg[i][0,:,:,k],cmap='jet')\n")
	f.write(""""        \n")
	f.write("gg2=fc.sess.run(fc.grd1,feed_dict=feed_dict)\n")
	f.write("gg3=fc.sess.run(fc.grd2,feed_dict=feed_dict)\n")
	f.write("\n")
	f.write("plt.figure()\n")
	f.write("plt.plot(gg[0][0,:,:,1].flatten(),color='blue',lw=8)\n")
	f.write("plt.plot(gg2[1][0][0].flatten(),color='red')\n")
	f.write("\n")
	f.write("print(len(gg2))\n")
	f.write("      \n")
	f.write("plt.figure()\n")
	f.write("for i in range(4):\n")
	f.write("    plt.subplot(2,2,1+i)\n")
	f.write("    plt.imshow(gg3[i][0][0].reshape(64,64),cmap='jet')\n")
	f.write("    \n")
	f.write("plt.figure()\n")
	f.write("print(g2[0].shape)\n")
	f.write("print(len(g2))\n")
	f.write("for i in range(4):\n")
	f.write("    plt.subplot(2,2,1+i)\n")
	f.write("    plt.imshow(g2[0][0,:,:,i],cmap='jet')\n")
	f.write("plt.show()\n")
	f.write("\n")
	f.write("exit(0)\n")
	f.write("\n")
	f.write("a=np.zeros([20])\n")
	f.write("b=np.zeros([20])\n")
	f.write("c=np.zeros([20])\n")
	f.write(""""\n")
	f.write("plt.figure()\n")
	f.write("for i in range(len(gg2)):\n")
	f.write("    plt.subplot(5,4,1+i)\n")
	f.write("    plt.imshow(gg2[i][0][0].reshape(64,64),cmap='jet')\n")
	f.write("    print((rr1[0,0,0,i]-or1[0,0,0,i])*1E6,gg[i//4][0,10//(2**(i//4)),10//(2**(i//4)),i%4],gg2[i][0][0][10+10*64])\n")
	f.write("    a[i]=(rr1[0,0,0,i]-or1[0,0,0,i])*1E6\n")
	f.write("    b[i]=gg[i//4][0,10//(2**(i//4)),10//(2**(i//4)),i%4]\n")
	f.write("    c[i]=gg2[i][0][0][10+10*64]\n")
	f.write("\n")
	f.write("plt.figure()\n")
	f.write("plt.plot(abs(b),color='red',lw=6)\n")
	f.write("plt.plot(abs(c),color='orange',lw=4)\n")
	f.write("plt.plot(abs(a),color='blue')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.show()\n")
	f.write(""""\n")
	f.write("\n")
	f.write("xidx={}\n")
	f.write("yidx={}\n")
	f.write("for i in range(5):\n")
	f.write("    xidx[i]=(np.arange(64*64,dtype='int')//64)//(2**i)\n")
	f.write("    yidx[i]=(np.arange(64*64,dtype='int')%64)//(2**i)\n")
	f.write("    \n")
	f.write("for itt in range(1000):\n")
	f.write("    imap=fc.get_map()\n")
	f.write("    imm=imap.reshape(1,64,64,1)\n")
	f.write("    rr1,rr2=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("    \n")
	f.write("    gg=fc.sess.run(grad,feed_dict=feed_dict)\n")
	f.write("    #gg2=fc.sess.run(fc.grd1,feed_dict=feed_dict)\n")
	f.write("\n")
	f.write("    diff=np.zeros([1,64,64,1])\n")
	f.write("    #diff2=np.zeros([1,64,64,1])\n")
	f.write("    """\n")
	f.write("    plt.figure(figsize=(16,8))\n")
	f.write("    """\n")
	f.write("    for i in range(5):\n")
	f.write("        tmp=np.sum(((rr1[0,0,0,i*4:(i+1)*4]-r1[0,0,0,i*4:(i+1)*4]).reshape(1,1,4))*(gg[i][0,:,:,:]/(64*64)),2)\n")
	f.write("        diff[0,:,:,0]+=(tmp[xidx[i],yidx[i]].reshape(64,64))/64\n")
	f.write("        """\n")
	f.write("        if i//4==0:\n")
	f.write("            tmp=(gg[i//4][0,:,:,i%4]/(64*64)).reshape(1,64,64,1)\n")
	f.write("        else:\n")
	f.write("            rtmp=(gg[i//4][0,:,:,i%4]/(64*64)).reshape(64//(2**(i//4)),64//(2**(i//4)))\n")
	f.write("\n")
	f.write("            tmp=np.zeros([64,64])\n")
	f.write("            xidx=np.arange(((64//(2**(i//4)))**2),dtype='int')//(64//(2**(i//4)))\n")
	f.write("            yidx=np.arange(((64//(2**(i//4)))**2),dtype='int')%(64//(2**(i//4)))\n")
	f.write("            for k in range(2**(i//4)):\n")
	f.write("                for l in range(2**(i//4)):\n")
	f.write("                    tmp[k+xidx*(2**(i//4)),l+yidx*(2**(i//4))]=rtmp[xidx,yidx]\n")
	f.write("    \n")
	f.write("        #diff+=((rr1[0,0,0,i]-r1[0,0,0,i])*gg2[i][0][0]/(64*64*64)).reshape(1,64,64,1)\n")
	f.write("        diff+=((rr1[0,0,0,i]-r1[0,0,0,i])*tmp/64.0).reshape(1,64,64,1)\n")
	f.write("        """\n")
	f.write("        """\n")
	f.write("        imm=imap.reshape(1,64,64,1)+((rr1[0,0,0,i]-r1[0,0,0,i])*gg2[i][0][0]/(16*64*64)).reshape(1,64,64,1)\n")
	f.write("        r1d,r2d=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("        \n")
	f.write("        imm=imap.reshape(1,64,64,1)+((rr1[0,0,0,i]-r1[0,0,0,i])*tmp/16).reshape(1,64,64,1)\n")
	f.write("        r1d2,r2d2=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("        \n")
	f.write("        plt.subplot(5,4,1+i)\n")
	f.write("        plt.plot(r1.flatten(),color='black')\n")
	f.write("        plt.plot(rr1.flatten(),color='blue')\n")
	f.write("        plt.plot(r1d.flatten(),color='orange')\n")
	f.write("        plt.plot(r1d2.flatten(),color='gray')\n")
	f.write("        plt.yscale('log')\n")
	f.write("        """\n")
	f.write("\n")
	f.write("    #imm=imap.reshape(1,64,64,1)+diff/2\n")
	f.write("    #r1d,r2d=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("    \n")
	f.write("    #imm=imap.reshape(1,64,64,1)+diff2/8\n")
	f.write("    #r1d2,r2d2=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("    \n")
	f.write("    par=fc.get_param()\n")
	f.write("    fc.set_value(par-(diff).flatten(),0)\n")
	f.write("    if itt%100==0:\n")
	f.write("        print(itt,(r1-rr1).std())\n")
	f.write("    """\n")
	f.write("    plt.figure()\n")
	f.write("    plt.plot(rr1n.flatten(),color='yellow',lw=8)\n")
	f.write("    plt.plot(r1.flatten(),color='black')\n")
	f.write("    plt.plot(rr1.flatten(),color='blue')\n")
	f.write("    plt.plot(r1d.flatten(),color='orange',lw=4)\n")
	f.write("    plt.plot(r1d2.flatten(),color='gray')\n")
	f.write("    plt.yscale('log')\n")
	f.write("    \n")
	f.write("\n")
	f.write("    plt.show()\n")
	f.write("    """\n")
	f.write("\n")
	f.write("plt.figure(figsize=(16,8))\n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.imshow(d.reshape(64,64),cmap='jet')\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.imshow(imap.reshape(64,64),cmap='jet')\n")
	f.write("plt.show()\n")
	f.write("exit(0)\n")
	f.write("for itt in range(10):\n")
	f.write("    gg2=fc.sess.run(fc.grd1,feed_dict=feed_dict)\n")
	f.write("    """\n")
	f.write("    plt.figure()\n")
	f.write("    for i in range(len(gg2)):\n")
	f.write("        plt.subplot(5,4,1+i)\n")
	f.write("        plt.imshow(gg2[i][0][0].reshape(64,64),cmap='jet')\n")
	f.write("    """\n")
	f.write("    imap=fc.get_map()\n")
	f.write("    \n")
	f.write("    imm=imap.reshape(1,64,64,1)\n")
	f.write("    rr1,rr2=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("    if itt==0:\n")
	f.write("        rrr1=rr1\n")
	f.write("\n")
	f.write("    mat=np.zeros([4096,4096])\n")
	f.write("    vec=np.zeros([4096])\n")
	f.write("    for i in range(20):\n")
	f.write("        imm=gg2[i][0][0].reshape(64*64)\n")
	f.write("        mat+=np.dot(imm.reshape(4096,1),imm.reshape(1,4096))\n")
	f.write("        vec+=(rr1[0,0,0,i]-r1[0,0,0,i])*imm/(20*64*64)*(2**(i//4)/16)\n")
	f.write("\n")
	f.write("    x=vec #conjugate(mat,vec)\n")
	f.write("\n")
	f.write("    imm=(imap+x.reshape(64,64)).reshape(1,64,64,1)\n")
	f.write("    r1d,r2d=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("    \n")
	f.write("    plt.figure()\n")
	f.write("    plt.plot(r1.flatten(),color='black')\n")
	f.write("    plt.plot(rrr1.flatten(),color='blue')\n")
	f.write("    plt.plot(r1d.flatten(),color='orange')\n")
	f.write("    plt.yscale('log')\n")
	f.write("    plt.show()\n")
	f.write("\n")
	f.write("    print(((rr1-r1)**2).sum())\n")
	f.write("    fc.set_value(-x,0)\n")
	f.write("\n")
	f.write("exit(0)\n")
	f.write("imm=(randfield+gg[0][0,:,:,1]).reshape(1,64,64,1)\n")
	f.write("r1d1,r2d1=fc.calc_stat(imm,imm,avg_ang=avg_ang)\n")
	f.write("\n")
	f.write("plt.plot(r1d1.flatten(),color='red')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.show()\n")
	f.write("exit(0)\n")
	f.write("# Run the learning\n")
	f.write("fc.learn(tw1,tw2,tb1,tb2,NUM_EPOCHS = 1000,DECAY_RATE=1.0)\n")
	f.write("\n")
	f.write("# get the output map\n")
	f.write("omap=fc.get_map()\n")
	f.write("\n")
	f.write("modd=omap.reshape(1,nx,nx)\n")
	f.write("o1,o2=fc.calc_stat(modd,modd,avg_ang=avg_ang)\n")
	f.write("#save the statistics on the synthesised data\n")
	f.write("np.save(scratch_path+'/%s_o1.npy'%(outname), o1)\n")
	f.write("np.save(scratch_path+'/%s_o2.npy'%(outname), o2)\n")
	f.write("\n")
	f.write("np.save(scratch_path+'/%s_ref.npy'%(outname), d)\n")
	f.write("np.save(scratch_path+'/%s_start.npy'%(outname), randfield)\n")
	f.write("np.save(scratch_path+'/%s_result.npy'%(outname),omap)\n")
	
	
	#=============================	
	
	f=open("testHealpix.py","w")

	f.write("import numpy as np\n")
	f.write("import os, sys\n")
	f.write("import matplotlib.pyplot as plt\n")
	f.write("import healpy as hp\n")
	f.write("import foscat.FoCUS as FOC\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# DEFINE A PATH FOR scratch data\n")
	f.write("# The data are storred using a default nside to minimize the needed storage\n")
	f.write("#=================================================================================\n")
	f.write("scratch_path = 'foscat_data'\n")
	f.write("outname='TEST'\n")
	f.write("Default_nside=512\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# INITIALIZE foscat class\n")
	f.write("#=================================================================================\n")
	f.write("fc=FOC.FoCUS(NORIENT=4,\n")
	f.write("             KERNELSZ=3,\n")
	f.write("             healpix=True,\n")
	f.write("             OSTEP=0,\n")
	f.write("             slope=1.0,\n")
	f.write("             TEMPLATE_PATH=scratch_path)\n")
	f.write("\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# DENOISE NSIDE=32 MAP\n")
	f.write("#=================================================================================\n")
	f.write("nout=32\n")
	f.write("#=================================================================================\n")
	f.write("# Function to reduce the data used in the foscat algorithm \n")
	f.write("#=================================================================================\n")
	f.write("def dodown(a,nout):\n")
	f.write("    nin=int(np.sqrt(a.shape[0]//12))\n")
	f.write("    if nin==nout:\n")
	f.write("        return(a)\n")
	f.write("    return(np.mean(a.reshape(12*nout*nout,(nin//nout)**2),1))\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Get data from web for demo\n")
	f.write("#=================================================================================\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Get dust simulated map from PySM2\n")
	f.write("#=================================================================================\n")
	f.write("try:\n")
	f.write("    d=dodown(np.load(scratch_path+'/dust_512.npy'),nout)\n")
	f.write("except:\n")
	f.write("    os.system('wget -O '+scratch_path+'/dust2comp_I1_ns512_545.fits https://portal.nersc.gov/project/cmb/pysm-data/pysm_2/dust2comp_I1_ns512_545.fits')\n")
	f.write("\n")
	f.write("    print('wget -O '+scratch_path+'/dust2comp_I1_ns512_545.fits https://portal.nersc.gov/project/cmb/pysm-data/pysm_2/dust2comp_I1_ns512_545.fits')\n")
	f.write("    im=hp.ud_grade(hp.read_map(scratch_path+'/dust2comp_I1_ns512_545.fits'),512)\n")
	f.write("    idx=hp.nest2ring(512,np.arange(12*512**2))\n")
	f.write("    im=10*im/im.std()\n")
	f.write("    np.save(scratch_path+'/dust_512.npy',im[idx])\n")
	f.write("    os.system('rm '+scratch_path+'/dust2comp_I1_ns512_545.fits')\n")
	f.write("    d=dodown(np.load(scratch_path+'/dust_512.npy'),nout)\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Get HI experimetal map\n")
	f.write("#=================================================================================\n")
	f.write("try:\n")
	f.write("    td=dodown(np.load(scratch_path+'/TH1_512.npy'),nout)\n")
	f.write("except:\n")
	f.write("    os.system('wget -O '+scratch_path+'/haslam408_dsds_Remazeilles2014.fits http://pla.esac.esa.int/pla/aio/product-action?MAP.MAP_ID=haslam408_dsds_Remazeilles2014.fits')\n")
	f.write("    h1=hp.ud_grade(hp.read_map(scratch_path+'/haslam408_dsds_Remazeilles2014.fits'),512)\n")
	f.write("    idx=hp.nest2ring(512,np.arange(12*512**2))\n")
	f.write("    h1=h1/h1.std()\n")
	f.write("    np.save(scratch_path+'/TH1_512.npy',h1[idx])\n")
	f.write("    os.system('rm '+scratch_path+'/haslam408_dsds_Remazeilles2014.fits')\n")
	f.write("    td=dodown(np.load(scratch_path+'/TH1_512.npy'),nout)\n")
	f.write("\n")
	f.write("# define the level of noise of the simulation\n")
	f.write("ampnoise=0.3\n")
	f.write("Alpha=0.8\n")
	f.write("nsim=100\n")
	f.write("avg_ang=False\n")
	f.write("\n")
	f.write("# Build to half mission by adding noise to the data on the first process\n")
	f.write("d1 = d+ampnoise*np.random.randn(12*nout**2)\n")
	f.write("d2 = d+ampnoise*np.random.randn(12*nout**2)\n")
	f.write("# simulate the noisy data\n")
	f.write("di = d+ampnoise*np.random.randn(12*nout**2)/np.sqrt(2)\n")
	f.write("np.save(scratch_path+'/test%sref.npy'%(outname), d)\n")
	f.write("np.save(scratch_path+'/test%sinput.npy'%(outname),di)\n")
	f.write("np.save(scratch_path+'/test%sinput1.npy'%(outname),d1)\n")
	f.write("np.save(scratch_path+'/test%sinput2.npy'%(outname),d2)\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# For real data:\n")
	f.write("# you have to give d1,d2 and d\n")
	f.write("# you have to define the value ampnoise\n")
	f.write("#=================================================================================\n")
	f.write("\n")
	f.write("\n")
	f.write("# All information of the map is used\n")
	f.write("nin=512\n")
	f.write("\n")
	f.write("tab=['MASK_GAL097_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL090_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL080_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL060_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL040_%d.npy'%(nin)]\n")
	f.write("\n")
	f.write("mask=np.ones([len(tab),12*nout**2])\n")
	f.write("for imask in range(len(tab)):\n")
	f.write("    try:\n")
	f.write("        mask[imask]=dodown(np.load(scratch_path+'/'+tab[imask]),nout)\n")
	f.write("    except:\n")
	f.write("        print('==========================================================')\n")
	f.write("        print('')\n")
	f.write("        print('Get offical Planck Galactic masks...may take few minutes..')\n")
	f.write("        print('')\n")
	f.write("        print('==========================================================')\n")
	f.write("\n")
	f.write("        os.system('wget -O '+scratch_path+'/HFI_Mask_GalPlane-apo5_2048_R2.00.fits https://irsa.ipac.caltech.edu/data/Planck/release_2/ancillary-data/masks/HFI_Mask_GalPlane-apo5_2048_R2.00.fits')\n")
	f.write("        tab_data=['GAL020','GAL040','GAL060','GAL070','GAL080','GAL090','GAL097','GAL099']\n")
	f.write("        \n")
	f.write("        idx=hp.nest2ring(512,np.arange(12*512**2))\n")
	f.write("        for i in range(8):\n")
	f.write("            lmask=hp.ud_grade(hp.read_map(scratch_path+'/HFI_Mask_GalPlane-apo5_2048_R2.00.fits',i),512)\n")
	f.write("            np.save(scratch_path+'/MASK_%s_512.npy'%(tab_data[i]),lmask[idx])\n")
	f.write("            print('Save '+scratch_path+'/MASK_%s_512.npy'%(tab_data[i]))\n")
	f.write("        os.system('rm '+scratch_path+'/HFI_Mask_GalPlane-apo5_2048_R2.00.fits')\n")
	f.write("        mask[imask]=dodown(np.load(scratch_path+'/'+tab[imask]),nout)\n")
	f.write("        \n")
	f.write("mask[0,:]=1.0\n")
	f.write("fc.add_mask(mask)\n")
	f.write("\n")
	f.write("# convert data in tensor for foscat (should be done internally)\n")
	f.write("data1 = fc.convimage(d1)\n")
	f.write("data2 = fc.convimage(d2)\n")
	f.write("data  = fc.convimage(di)\n")
	f.write("tdata = fc.convimage(td)\n")
	f.write("\n")
	f.write("# Initialize the learning and initialize the tensor to be synthesized\n")
	f.write("ldata=fc.init_synthese(di)\n")
	f.write("\n")
	f.write("# Add losss:\n")
	f.write("# here d1 x d2 = s x s\n")
	f.write("fc.add_loss_healpix(data1,data2,ldata,ldata,avg_ang=avg_ang)\n")
	f.write("# here d1 x d2 = d x s\n")
	f.write("fc.add_loss_healpix(data,ldata,ldata,ldata,avg_ang=avg_ang)\n")
	f.write("# here T x d = T x s\n")
	f.write("fc.add_loss_healpix(tdata,data,tdata,ldata,avg_ang=avg_ang,imaginary=True)\n")
	f.write("\n")
	f.write("# initiliaze the loss\n")
	f.write("loss=fc.init_optim()\n")
	f.write("\n")
	f.write("moddr=d.reshape(1,12*nout**2)\n")
	f.write("x1,x2=fc.calc_stat(moddr,moddr,avg_ang=avg_ang)\n")
	f.write("np.save(scratch_path+'/i%s1_%d.npy'%(outname,0), x1)\n")
	f.write("np.save(scratch_path+'/i%s2_%d.npy'%(outname,0), x2)\n")
	f.write("\n")
	f.write("c1,c2=fc.calc_stat(d1.reshape(1,12*nout**2),d2.reshape(1,12*nout**2),avg_ang=avg_ang)\n")
	f.write("np.save(scratch_path+'/c%s1_%d.npy'%(outname,0), c1)\n")
	f.write("np.save(scratch_path+'/c%s2_%d.npy'%(outname,0), c2)\n")
	f.write("lmap1=1*d1\n")
	f.write("lmap2=1*d2\n")
	f.write("\n")
	f.write("tw1={}\n")
	f.write("tw2={}\n")
	f.write("tb1={}\n")
	f.write("tb2={}\n")
	f.write("for i in range(3):\n")
	f.write("    tw1[i]=0.0\n")
	f.write("    tw2[i]=0.0\n")
	f.write("    tb1[i]=0.0\n")
	f.write("    tb2[i]=0.0\n")
	f.write("    \n")
	f.write("for itt in range(5):\n")
	f.write("    # compute scattering to make weights\n")
	f.write("    modd1=lmap1.reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("    modd2=lmap2.reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("    o1,o2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=0)\n")
	f.write("    modd1=lmap1.reshape(1,12*nout**2)\n")
	f.write("    modd2=lmap2.reshape(1,12*nout**2)\n")
	f.write("    r1,r2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=0)\n")
	f.write("    #calcul poids des coeffs\n")
	f.write("    np.save(scratch_path+'/b%s1_%d.npy'%(outname,itt), r1)\n")
	f.write("    np.save(scratch_path+'/b%s2_%d.npy'%(outname,itt), r2)\n")
	f.write("    np.save(scratch_path+'/n%s1_%d.npy'%(outname,itt), o1)\n")
	f.write("    np.save(scratch_path+'/n%s2_%d.npy'%(outname,itt), o2)\n")
	f.write("        \n")
	f.write("    modd1=lmap1.reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("    modd2=lmap2.reshape(1,12*nout**2)+0*np.random.randn(nsim,12*nout**2)\n")
	f.write("    of1,of2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=1)\n")
	f.write("    \n")
	f.write("    modd1=td.reshape(1,12*nout**2)+0*np.random.randn(nsim,12*nout**2)\n")
	f.write("    modd2=((lmap1+lmap2)/2).reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("    onx1,onx2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,imaginary=True,gpupos=2)\n")
	f.write("    modd1=td.reshape(1,12*nout**2)\n")
	f.write("    modd2=((lmap1+lmap2)/2).reshape(1,12*nout**2)\n")
	f.write("    ox1,ox2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,imaginary=True,gpupos=2)\n")
	f.write("    \n")
	f.write("    tw1[0]=1/np.std(o1,0)\n")
	f.write("    tw2[0]=1/np.std(o2,0)\n")
	f.write("    tw1[1]=1/np.std(of1,0)\n")
	f.write("    tw2[1]=1/np.std(of2,0)\n")
	f.write("    tw1[2]=1/np.std(onx1,0)\n")
	f.write("    tw2[2]=1/np.std(onx2,0)\n")
	f.write("    \n")
	f.write("    tb1[0]=Alpha*(np.mean(o1-r1,0)-tb1[0])+tb1[0]\n")
	f.write("    tb2[0]=Alpha*(np.mean(o2-r2,0)-tb2[0])+tb2[0]\n")
	f.write("    tb1[1]=Alpha*(np.mean(of1-r1,0)-tb1[1])+tb1[1]\n")
	f.write("    tb2[1]=Alpha*(np.mean(of2-r2,0)-tb2[1])+tb2[1]\n")
	f.write("    tb1[2]=Alpha*(np.mean(onx1-ox1,0)-tb1[2])+tb1[2]\n")
	f.write("    tb2[2]=Alpha*(np.mean(onx2-ox2,0)-tb2[2])+tb2[2]\n")
	f.write("    \n")
	f.write("    # make the learn\n")
	f.write("    fc.reset()\n")
	f.write("    \n")
	f.write("    omap=fc.learn(tw1,tw2,tb1,tb2,NUM_EPOCHS = 500,DECAY_RATE=0.995,LEARNING_RATE=0.03,ADDAPT_LEARN=2.0)\n")
	f.write("\n")
	f.write("    print('ITT ',itt,((d-omap)*mask[1].reshape(12*nout**2)).std(),((d-di)*mask[1].reshape(12*nout**2)).std())\n")
	f.write("    sys.stdout.flush()\n")
	f.write("    modd1=omap.reshape(1,12*nout**2)\n")
	f.write("    oo1,oo2=fc.calc_stat(modd1,modd1,avg_ang=avg_ang)\n")
	f.write("    lmap1=1*omap\n")
	f.write("    lmap2=1*omap\n")
	f.write("    np.save(scratch_path+'/o%s1_%d.npy'%(outname,itt), oo1)\n")
	f.write("    np.save(scratch_path+'/o%s2_%d.npy'%(outname,itt), oo2)\n")
	f.write("    np.save(scratch_path+'/test%sresult_%d.npy'%(outname,itt),omap)\n")
	f.write("\n")
	f.write("\n")
	
	
	#=============================	
	
	f=open("testHplot.py","w")

	f.write("import numpy as np\n")
	f.write("import os, sys\n")
	f.write("import healpy as hp\n")
	f.write("import matplotlib.pyplot as plt\n")
	f.write("\n")
	f.write("nitt=5\n")
	f.write("scratch_path = 'foscat_data'\n")
	f.write("outname='TEST'\n")
	f.write("\n")
	f.write("def dodown(a,nout):\n")
	f.write("    nin=int(np.sqrt(a.shape[0]//12))\n")
	f.write("    if nin==nout:\n")
	f.write("        return(a)\n")
	f.write("    return(np.mean(a.reshape(12*nout*nout,(nin//nout)**2),1))\n")
	f.write("\n")
	f.write("plt.figure(figsize=(12,12))\n")
	f.write("#reference\n")
	f.write("i1=np.load(scratch_path+'/i%s1_%d.npy'%(outname,0))\n")
	f.write("i2=np.load(scratch_path+'/i%s2_%d.npy'%(outname,0))\n")
	f.write("#mesure\n")
	f.write("c1=np.load(scratch_path+'/c%s1_%d.npy'%(outname,0))\n")
	f.write("c2=np.load(scratch_path+'/c%s2_%d.npy'%(outname,0))\n")
	f.write("\n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.plot(abs(1-i1.flatten()/i1.flatten()),color='blue')\n")
	f.write("plt.plot(abs(1-c1.flatten()/i1.flatten()),color='orange')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.ylabel(r'$S_1$')\n")
	f.write("plt.xlabel('j1')\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.plot(abs(1-i2.flatten()/i2.flatten()),color='blue')\n")
	f.write("plt.plot(abs(1-c2.flatten()/i2.flatten()),color='orange')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.ylabel(r'$S_2$')\n")
	f.write("plt.xlabel('j1,j2')\n")
	f.write("    \n")
	f.write("for itt in range(nitt):\n")
	f.write("    #foscat map\n")
	f.write("    o1=np.load(scratch_path+'/o%s1_%d.npy'%(outname,itt))\n")
	f.write("    o2=np.load(scratch_path+'/o%s2_%d.npy'%(outname,itt))\n")
	f.write("\n")
	f.write("    plt.subplot(1,2,1)\n")
	f.write("    plt.plot(abs(1-o1.flatten()/i1.flatten()),color='red',lw=1)\n")
	f.write("    plt.subplot(1,2,2)\n")
	f.write("    plt.plot(abs(1-o2.flatten()/i2.flatten()),color='red',lw=1)\n")
	f.write("    \n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.plot(abs(1-o1.flatten()/i1.flatten()),color='red',lw=4)\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.plot(abs(1-o2.flatten()/i2.flatten()),color='red',lw=4)\n")
	f.write("    \n")
	f.write("plt.subplot(1,2,1)\n")
	f.write("plt.plot(abs(1-i1.flatten()/i1.flatten()),color='blue')\n")
	f.write("plt.plot(abs(1-c1.flatten()/i1.flatten()),color='orange')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.ylabel(r'$S_1$')\n")
	f.write("plt.xlabel('j1')\n")
	f.write("plt.subplot(1,2,2)\n")
	f.write("plt.plot(abs(1-i2.flatten()/i2.flatten()),color='blue')\n")
	f.write("plt.plot(abs(1-c2.flatten()/i2.flatten()),color='orange')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.ylabel(r'$S_2$')\n")
	f.write("plt.xlabel('j1,j2')\n")
	f.write("\n")
	f.write("d  = np.load(scratch_path+'/test%sref.npy'%(outname))\n")
	f.write("d1 = np.load(scratch_path+'/test%sinput1.npy'%(outname))\n")
	f.write("d2 = np.load(scratch_path+'/test%sinput2.npy'%(outname))\n")
	f.write("di = np.load(scratch_path+'/test%sinput.npy'%(outname))\n")
	f.write("s  = np.load(scratch_path+'/test%sresult_%d.npy'%(outname,nitt-1)).flatten()\n")
	f.write("\n")
	f.write("amp=4\n")
	f.write("\n")
	f.write("plt.figure(figsize=(12,6))\n")
	f.write("hp.mollview(d,cmap='jet',    nest=True,hold=False,sub=(2,3,1),min=-amp,max=amp,title='Model',norm='hist')\n")
	f.write("hp.mollview(di,cmap='jet',   nest=True,hold=False,sub=(2,3,2),min=-amp,max=amp,title='Noisy',norm='hist')\n")
	f.write("hp.mollview(s,cmap='jet',    nest=True,hold=False,sub=(2,3,3),min=-amp,max=amp,title='Cleanned',norm='hist')\n")
	f.write("hp.mollview(di-d,cmap='jet', nest=True,hold=False,sub=(2,3,4),min=-amp/4,max=amp/4,title='Noisy-Model')\n")
	f.write("hp.mollview(di-s,cmap='jet', nest=True,hold=False,sub=(2,3,5),min=-amp/4,max=amp/4,title='Noisy-Cleanned')\n")
	f.write("hp.mollview(s-d,cmap='jet',  nest=True,hold=False,sub=(2,3,6),min=-amp/4,max=amp/4,title='Cleanned-Model')\n")
	f.write("\n")
	f.write("nin=512\n")
	f.write("nout=int(np.sqrt(d.shape[0]//12))\n")
	f.write("idx=hp.ring2nest(nout,np.arange(12*nout**2))\n")
	f.write("\n")
	f.write("mask=dodown(np.load(scratch_path+'/MASK_GAL080_%d.npy'%(nin)),nout)\n")
	f.write("clr=hp.anafast((mask*d-np.median(mask*d))[idx])\n")
	f.write("cli=hp.anafast((mask*di-np.median(mask*di))[idx])\n")
	f.write("cln=hp.anafast((mask*(d-di)-np.median(mask*(d-di)))[idx])\n")
	f.write("\n")
	f.write("plt.figure(figsize=(12,12))\n")
	f.write("\n")
	f.write("for itt in range(nitt):\n")
	f.write("    s  = np.load(scratch_path+'/test%sresult_%d.npy'%(outname,itt)).flatten()\n")
	f.write("\n")
	f.write("    clo=hp.anafast((mask*s-np.median(mask*s))[idx])\n")
	f.write("    cld=hp.anafast((mask*(d-s)-np.median(mask*(d-s)))[idx])\n")
	f.write("\n")
	f.write("    plt.plot(clo,color='orange',label='s %d'%(itt),lw=1)\n")
	f.write("    plt.plot(cld,color='red',label='r-s %d'%(itt),lw=1)\n")
	f.write("\n")
	f.write("plt.plot(clo,color='orange',label='s %d'%(itt),lw=4)\n")
	f.write("plt.plot(cld,color='red',label='r-s %d'%(itt),lw=4)\n")
	f.write("    \n")
	f.write("plt.plot(clr,color='black',label='r')\n")
	f.write("plt.plot(cli,color='blue',label='d')\n")
	f.write("plt.plot(cln,color='grey',label='r-d')\n")
	f.write("\n")
	f.write("plt.legend()\n")
	f.write("plt.xscale('log')\n")
	f.write("plt.yscale('log')\n")
	f.write("plt.show()\n")
	
	
	#=============================	
	
	f=open("testHealpix_mpi.py","w")

	f.write("import numpy as np\n")
	f.write("import os, sys\n")
	f.write("import matplotlib.pyplot as plt\n")
	f.write("import healpy as hp\n")
	f.write("import foscat.FoCUS as FOC\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# DEFINE A PATH FOR scratch data\n")
	f.write("# The data are storred using a default nside to minimize the needed storage\n")
	f.write("#=================================================================================\n")
	f.write("scratch_path = 'foscat_data'\n")
	f.write("outname='TEST_MPI'\n")
	f.write("Default_nside=512\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# INITIALIZE foscat class\n")
	f.write("#=================================================================================\n")
	f.write("fc=FOC.FoCUS(NORIENT=8,\n")
	f.write("              KERNELSZ=5,\n")
	f.write("              healpix=True,\n")
	f.write("              OSTEP=0,\n")
	f.write("              slope=1.0,\n")
	f.write("              isMPI=True,\n")
	f.write("              TEMPLATE_PATH=scratch_path)\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# DENOISE NSIDE=32 MAP\n")
	f.write("#=================================================================================\n")
	f.write("nout=32\n")
	f.write("#=================================================================================\n")
	f.write("# Function to reduce the data used in the foscat algorithm \n")
	f.write("#=================================================================================\n")
	f.write("def dodown(a,nout):\n")
	f.write("    nin=int(np.sqrt(a.shape[0]//12))\n")
	f.write("    if nin==nout:\n")
	f.write("        return(a)\n")
	f.write("    return(np.mean(a.reshape(12*nout*nout,(nin//nout)**2),1))\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Get data from web for demo\n")
	f.write("#=================================================================================\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Get dust simulated map from PySM2\n")
	f.write("#=================================================================================\n")
	f.write("try:\n")
	f.write("    d=dodown(np.load(scratch_path+'/dust_512.npy'),nout)\n")
	f.write("except:\n")
	f.write("    if fc.get_rank()==0:\n")
	f.write("        os.system('wget -O '+scratch_path+'/dust2comp_I1_ns512_545.fits https://portal.nersc.gov/project/cmb/pysm-data/pysm_2/dust2comp_I1_ns512_545.fits')\n")
	f.write("\n")
	f.write("        print('wget -O '+scratch_path+'/dust2comp_I1_ns512_545.fits https://portal.nersc.gov/project/cmb/pysm-data/pysm_2/dust2comp_I1_ns512_545.fits')\n")
	f.write("        im=hp.ud_grade(hp.read_map(scratch_path+'/dust2comp_I1_ns512_545.fits'),512)\n")
	f.write("        idx=hp.nest2ring(512,np.arange(12*512**2))\n")
	f.write("        im=10*im/im.std()\n")
	f.write("        np.save(scratch_path+'/dust_512.npy',im[idx])\n")
	f.write("        os.system('rm '+scratch_path+'/dust2comp_I1_ns512_545.fits')\n")
	f.write("    fc.barrier()\n")
	f.write("    \n")
	f.write("    d=dodown(np.load(scratch_path+'/dust_512.npy'),nout)\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# Get HI experimetal map\n")
	f.write("#=================================================================================\n")
	f.write("try:\n")
	f.write("    td=dodown(np.load(scratch_path+'/TH1_512.npy'),nout)\n")
	f.write("except:\n")
	f.write("    if fc.get_rank()==0:\n")
	f.write("        os.system('wget -O '+scratch_path+'/haslam408_dsds_Remazeilles2014.fits http://pla.esac.esa.int/pla/aio/product-action?MAP.MAP_ID=haslam408_dsds_Remazeilles2014.fits')\n")
	f.write("        h1=hp.ud_grade(hp.read_map(scratch_path+'/haslam408_dsds_Remazeilles2014.fits'),512)\n")
	f.write("        idx=hp.nest2ring(512,np.arange(12*512**2))\n")
	f.write("        h1=h1/h1.std()\n")
	f.write("        np.save(scratch_path+'/TH1_512.npy',h1[idx])\n")
	f.write("        os.system('rm '+scratch_path+'/haslam408_dsds_Remazeilles2014.fits')\n")
	f.write("    fc.barrier()\n")
	f.write("    td=dodown(np.load(scratch_path+'/TH1_512.npy'),nout)\n")
	f.write("\n")
	f.write("\n")
	f.write("# define the level of noise of the simulation\n")
	f.write("ampnoise=0.3\n")
	f.write("Alpha=1.0\n")
	f.write("nsim=100\n")
	f.write("avg_ang=True\n")
	f.write("\n")
	f.write("# Build to half mission by adding noise to the data on the first process\n")
	f.write("if fc.get_rank()==0:\n")
	f.write("    d1 = d+ampnoise*np.random.randn(12*nout**2)\n")
	f.write("    d2 = d+ampnoise*np.random.randn(12*nout**2)\n")
	f.write("    # simulate the noisy data\n")
	f.write("    di = d+ampnoise*np.random.randn(12*nout**2)/np.sqrt(2)\n")
	f.write("    np.save(scratch_path+'/test%sref.npy'%(outname), d)\n")
	f.write("    np.save(scratch_path+'/test%sinput.npy'%(outname),di)\n")
	f.write("    np.save(scratch_path+'/test%sinput1.npy'%(outname),d1)\n")
	f.write("    np.save(scratch_path+'/test%sinput2.npy'%(outname),d2)\n")
	f.write("    \n")
	f.write("# Ensure to use the same input data into all process\n")
	f.write("fc.barrier()\n")
	f.write("\n")
	f.write("d=np.load(scratch_path+'/test%sref.npy'%(outname))\n")
	f.write("di=np.load(scratch_path+'/test%sinput.npy'%(outname))\n")
	f.write("d1=np.load(scratch_path+'/test%sinput1.npy'%(outname))\n")
	f.write("d2=np.load(scratch_path+'/test%sinput2.npy'%(outname))\n")
	f.write("\n")
	f.write("#=================================================================================\n")
	f.write("# For real data:\n")
	f.write("# you have to give d1,d2 and d\n")
	f.write("# you have to define the value ampnoise\n")
	f.write("#=================================================================================\n")
	f.write("\n")
	f.write("\n")
	f.write("# All information of the map is used\n")
	f.write("nin=512\n")
	f.write("\n")
	f.write("tab=['MASK_GAL097_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL090_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL080_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL060_%d.npy'%(nin),\n")
	f.write("     'MASK_GAL040_%d.npy'%(nin)]\n")
	f.write("\n")
	f.write("mask=np.ones([len(tab),12*nout**2])\n")
	f.write("for imask in range(len(tab)):\n")
	f.write("    try:\n")
	f.write("        mask[imask]=dodown(np.load(scratch_path+'/'+tab[imask]),nout)\n")
	f.write("    except:\n")
	f.write("        if fc.get_rank()==0:\n")
	f.write("            print('==========================================================')\n")
	f.write("            print('')\n")
	f.write("            print('Get offical Planck Galactic masks...may take few minutes..')\n")
	f.write("            print('')\n")
	f.write("            print('==========================================================')\n")
	f.write("\n")
	f.write("            os.system('wget -O '+scratch_path+'/HFI_Mask_GalPlane-apo5_2048_R2.00.fits https://irsa.ipac.caltech.edu/data/Planck/release_2/ancillary-data/masks/HFI_Mask_GalPlane-apo5_2048_R2.00.fits')\n")
	f.write("            tab_data=['GAL020','GAL040','GAL060','GAL070','GAL080','GAL090','GAL097','GAL099']\n")
	f.write("\n")
	f.write("            idx=hp.nest2ring(512,np.arange(12*512**2))\n")
	f.write("            for i in range(8):\n")
	f.write("                lmask=hp.ud_grade(hp.read_map(scratch_path+'/HFI_Mask_GalPlane-apo5_2048_R2.00.fits',i),512)\n")
	f.write("                np.save(scratch_path+'/MASK_%s_512.npy'%(tab_data[i]),lmask[idx])\n")
	f.write("                print('Save '+scratch_path+'/MASK_%s_512.npy'%(tab_data[i]))\n")
	f.write("            os.system('rm '+scratch_path+'/HFI_Mask_GalPlane-apo5_2048_R2.00.fits')\n")
	f.write("        fc.barrier()\n")
	f.write("        mask[imask]=dodown(np.load(scratch_path+'/'+tab[imask]),nout)\n")
	f.write("        \n")
	f.write("mask[0,:]=1.0\n")
	f.write("fc.add_mask(mask)\n")
	f.write("\n")
	f.write("# convert data in tensor for foscat (should be done internally)\n")
	f.write("data1 = fc.convimage(d1)\n")
	f.write("data2 = fc.convimage(d2)\n")
	f.write("data  = fc.convimage(di)\n")
	f.write("tdata = fc.convimage(td)\n")
	f.write("\n")
	f.write("# Initialize the learning and initialize the tensor to be synthesized\n")
	f.write("ldata=fc.init_synthese(di)\n")
	f.write("\n")
	f.write("# Add losss:\n")
	f.write("# here d1 x d2 = s x s\n")
	f.write("if fc.get_rank()==0:\n")
	f.write("    fc.add_loss_healpix(data1,data2,ldata,ldata,avg_ang=avg_ang)\n")
	f.write("# here d1 x d2 = d x s\n")
	f.write("if fc.get_rank()==1:\n")
	f.write("    fc.add_loss_healpix(data,ldata,ldata,ldata,avg_ang=avg_ang)\n")
	f.write("# here T x d = T x s\n")
	f.write("if fc.get_rank()==2:\n")
	f.write("    fc.add_loss_healpix(tdata,data,tdata,ldata,avg_ang=avg_ang,imaginary=True)\n")
	f.write("\n")
	f.write("# initiliaze the loss\n")
	f.write("loss=fc.init_optim()\n")
	f.write("\n")
	f.write("moddr=d.reshape(1,12*nout**2)\n")
	f.write("x1,x2=fc.calc_stat(moddr,moddr,avg_ang=avg_ang)\n")
	f.write("if fc.get_rank()==0:\n")
	f.write("    np.save(scratch_path+'/i%s1_%d.npy'%(outname,0), x1)\n")
	f.write("    np.save(scratch_path+'/i%s2_%d.npy'%(outname,0), x2)\n")
	f.write("\n")
	f.write("c1,c2=fc.calc_stat(d1.reshape(1,12*nout**2),d2.reshape(1,12*nout**2),avg_ang=avg_ang)\n")
	f.write("if fc.get_rank()==0:\n")
	f.write("    np.save(scratch_path+'/c%s1_%d.npy'%(outname,0), c1)\n")
	f.write("    np.save(scratch_path+'/c%s2_%d.npy'%(outname,0), c2)\n")
	f.write("lmap1=1*d1\n")
	f.write("lmap2=1*d2\n")
	f.write("\n")
	f.write("tw1={}\n")
	f.write("tw2={}\n")
	f.write("tb1={}\n")
	f.write("tb2={}\n")
	f.write("for i in range(3):\n")
	f.write("    tw1[i]=0.0\n")
	f.write("    tw2[i]=0.0\n")
	f.write("    tb1[i]=0.0\n")
	f.write("    tb2[i]=0.0\n")
	f.write("\n")
	f.write("for itt in range(10):\n")
	f.write("    if fc.get_rank()==0:\n")
	f.write("        # compute scattering to make weights\n")
	f.write("        modd1=lmap1.reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("        modd2=lmap2.reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("        o1,o2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=0)\n")
	f.write("        modd1=lmap1.reshape(1,12*nout**2)\n")
	f.write("        modd2=lmap2.reshape(1,12*nout**2)\n")
	f.write("        r1,r2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=0)\n")
	f.write("        #calcul poids des coeffs\n")
	f.write("        np.save(scratch_path+'/b%s1_%d.npy'%(outname,itt), r1)\n")
	f.write("        np.save(scratch_path+'/b%s2_%d.npy'%(outname,itt), r2)\n")
	f.write("        np.save(scratch_path+'/n%s1_%d.npy'%(outname,itt), o1)\n")
	f.write("        np.save(scratch_path+'/n%s2_%d.npy'%(outname,itt), o2)\n")
	f.write("        \n")
	f.write("    if fc.get_rank()==1:\n")
	f.write("        modd1=lmap1.reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("        modd2=lmap2.reshape(1,12*nout**2)+0*np.random.randn(nsim,12*nout**2)\n")
	f.write("        of1,of2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=1)\n")
	f.write("        modd1=lmap1.reshape(1,12*nout**2)\n")
	f.write("        modd2=lmap2.reshape(1,12*nout**2)\n")
	f.write("        r1,r2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,gpupos=1)\n")
	f.write("    if fc.get_rank()==2:\n")
	f.write("        modd1=td.reshape(1,12*nout**2)+0*np.random.randn(nsim,12*nout**2)\n")
	f.write("        modd2=((lmap1+lmap2)/2).reshape(1,12*nout**2)+ampnoise*np.random.randn(nsim,12*nout**2)\n")
	f.write("        onx1,onx2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,imaginary=True,gpupos=2)\n")
	f.write("        modd1=td.reshape(1,12*nout**2)\n")
	f.write("        modd2=((lmap1+lmap2)/2).reshape(1,12*nout**2)\n")
	f.write("        ox1,ox2=fc.calc_stat(modd1,modd2,avg_ang=avg_ang,imaginary=True,gpupos=2)\n")
	f.write("    \n")
	f.write("    \n")
	f.write("    if fc.get_rank()==0:\n")
	f.write("        tw1[0]=1/np.std(o1,0)\n")
	f.write("        tw2[0]=1/np.std(o2,0)\n")
	f.write("    if fc.get_rank()==1:\n")
	f.write("        tw1[0]=1/np.std(of1,0)\n")
	f.write("        tw2[0]=1/np.std(of2,0)\n")
	f.write("    if fc.get_rank()==2:\n")
	f.write("        tw1[0]=1/np.std(onx1,0)\n")
	f.write("        tw2[0]=1/np.std(onx2,0)\n")
	f.write("    \n")
	f.write("    if fc.get_rank()==0:\n")
	f.write("        tb1[0]=Alpha*(np.mean(o1-r1,0)-tb1[0])+tb1[0]\n")
	f.write("        tb2[0]=Alpha*(np.mean(o2-r2,0)-tb2[0])+tb2[0]\n")
	f.write("    if fc.get_rank()==1:\n")
	f.write("        tb1[0]=Alpha*(np.mean(of1-r1,0)-tb1[0])+tb1[0]\n")
	f.write("        tb2[0]=Alpha*(np.mean(of2-r2,0)-tb2[0])+tb2[0]\n")
	f.write("    if fc.get_rank()==2:\n")
	f.write("        tb1[0]=Alpha*(np.mean(onx1-ox1,0)-tb1[0])+tb1[0]\n")
	f.write("        tb2[0]=Alpha*(np.mean(onx2-ox2,0)-tb2[0])+tb2[0]\n")
	f.write("    \n")
	f.write("    # make the learn\n")
	f.write("    fc.reset()\n")
	f.write("    \n")
	f.write("    omap=fc.learn(tw1,tw2,tb1,tb2,NUM_EPOCHS = 1000,DECAY_RATE=0.995,LEARNING_RATE=0.03,ADDAPT_LEARN=2.0)\n")
	f.write("\n")
	f.write("    if fc.get_rank()==0:\n")
	f.write("        print('ITT ',itt,((d-omap)*mask[1].reshape(12*nout**2)).std(),((d-di)*mask[1].reshape(12*nout**2)).std())\n")
	f.write("        sys.stdout.flush()\n")
	f.write("        modd1=omap.reshape(1,12*nout**2)\n")
	f.write("        oo1,oo2=fc.calc_stat(modd1,modd1,avg_ang=avg_ang)\n")
	f.write("        lmap1=1*omap\n")
	f.write("        lmap2=1*omap\n")
	f.write("        np.save(scratch_path+'/o%s1_%d.npy'%(outname,itt), oo1)\n")
	f.write("        np.save(scratch_path+'/o%s2_%d.npy'%(outname,itt), oo2)\n")
	f.write("        np.save(scratch_path+'/test%sresult_%d.npy'%(outname,itt),omap)\n")
	f.write("\n")
	f.write("\n")
	f.close()	
#======= END OF DEMO ========



import numpy as np

class Rformat:
    def __init__(self,
                 im,
                 off,
                 axis,
                 chans=12):
        self.data=im
        self.shape=im.shape
        self.axis=axis
        self.off=off
        self.chans=chans
        self.nside=im.shape[axis+1]-2*off

    def get(self):
        return self.data
    
    def get_data(self):
        if self.axis==0:
            return self.data[:,self.off:-self.off,self.off:-self.off]
        if self.axis==1:
            return self.data[:,:,self.off:-self.off,self.off:-self.off]
        if self.axis==2:
            return self.data[:,:,:,self.off:-self.off,self.off:-self.off]
        if self.axis==3:
            return self.data[:,:,:,:,self.off:-self.off,self.off:-self.off]

        print('get_data is implemented till axis==3 and axis=',self.axis)
        exit(0)

    def __add__(self,other):
        
        if isinstance(other, Rformat):
            return Rformat(self.get()+other.get(),self.off,self.axis,chans=self.chans)
        else:
            return Rformat(self.get()+other,self.off,self.axis,chans=self.chans)

    def __sub__(self,other):
        
        if isinstance(other, Rformat):
            return Rformat(self.get()-other.get(),self.off,self.axis,chans=self.chans)
        else:
            return Rformat(self.get()-other,self.off,self.axis,chans=self.chans)
            
    def __neg__(self):
        
        return Rformat(-self.get(),self.off,self.axis,chans=self.chans)

    def __mul__(self,other):
        
        if isinstance(other, Rformat):
            return Rformat(self.get()*other.get(),self.off,self.axis,chans=self.chans)
        else:
            return Rformat(self.get()*other,self.off,self.axis,chans=self.chans)



import sys
import numpy as np
import foscat.Rformat as Rformat

class foscat_backend:
    
    def __init__(self,name,mpi_rank=0,all_type='float64',gpupos=0):
        
        self.TENSORFLOW=1
        self.TORCH=2
        self.NUMPY=3
        
        self.BACKEND=name
        
        if name not in ['tensorflow','torch','numpy']:
            print('Backend %s not yet implemented')
            exit(0)
            
        if self.BACKEND=='tensorflow':
            import tensorflow as tf
            
            self.backend=tf
            self.BACKEND=self.TENSORFLOW
            #tf.config.threading.set_inter_op_parallelism_threads(1)
            #tf.config.threading.set_intra_op_parallelism_threads(1)

        if self.BACKEND=='torch':
            import torch
            self.BACKEND=self.TORCH
            self.backend=torch
            
        if self.BACKEND=='numpy':
            self.BACKEND=self.NUMPY
            self.backend=np
            
        self.float64=self.backend.float64
        self.float32=self.backend.float32
        self.int64=self.backend.int64
        self.int32=self.backend.int32
        self.complex64=self.backend.complex128
        self.complex128=self.backend.complex64
        
        if all_type=='float32':
            self.all_bk_type=self.backend.float32
            self.all_cbk_type=self.backend.complex64
        else:
            if all_type=='float64':
                self.all_type='float64'
                self.all_bk_type=self.backend.float64
                self.all_cbk_type=self.backend.complex128
            else:
                print('ERROR INIT FOCUS ',all_type,' should be float32 or float64')
                exit(0)
        #===========================================================================
        # INIT 
        if mpi_rank==0:
            if self.BACKEND==self.TENSORFLOW:
                print("Num GPUs Available: ", len(self.backend.config.experimental.list_physical_devices('GPU')))
            sys.stdout.flush()
        
        if self.BACKEND==self.TENSORFLOW:
            self.backend.debugging.set_log_device_placement(False)
            self.backend.config.set_soft_device_placement(True)
        
            gpus = self.backend.config.experimental.list_physical_devices('GPU')
            
        if self.BACKEND==self.TORCH:
            gpus=torch.cuda.is_available()
            
        gpuname='CPU:0'
        self.gpulist={}
        self.gpulist[0]=gpuname
        self.ngpu=1
        
        if gpus:
            try:
                if self.BACKEND==self.TENSORFLOW:
                # Currently, memory growth needs to be the same across GPUs
                    for gpu in gpus:
                        self.backend.config.experimental.set_memory_growth(gpu, True)
                    logical_gpus = self.backend.config.experimental.list_logical_devices('GPU')
                    print(len(gpus), "Physical GPUs,", len(logical_gpus), "Logical GPUs")
                    sys.stdout.flush()
                    self.ngpu=len(logical_gpus)
                    gpuname=logical_gpus[gpupos%self.ngpu].name
                    self.gpulist={}
                    for i in range(self.ngpu):
                        self.gpulist[i]=logical_gpus[i].name
                if self.BACKEND==self.TORCH:
                    self.ngpu=torch.cuda.device_count()
                    self.gpulist={}
                    for k in range(self.ngpu):
                        self.gpulist[k]=torch.cuda.get_device_name(0)

            except RuntimeError as e:
                # Memory growth must be set before GPUs have been initialized
                print(e)

        
    # ---------------------------------------------−---------
    # --             BACKEND DEFINITION                    --
    # ---------------------------------------------−---------
    def conv2d(self,x,w,strides=[1, 1, 1, 1],padding='SAME'):
        if self.BACKEND==self.TENSORFLOW:
                return self.backend.nn.conv2d(x,w,
                                               strides=strides,
                                               padding=padding)
        # to be written!!!
        if self.BACKEND==self.TORCH:
            return x
        if self.BACKEND==self.NUMPY:
            return x
        
    def bk_device(self,device_name):
        return self.backend.device(device_name)
        
    def bk_ones(self,shape,dtype=None):
        if dtype is None:
            dtype=self.all_type
        return(self.backend.ones(shape,dtype=dtype))

    def bk_L1(self,x):
        if isinstance(x,Rformat.Rformat):
            return Rformat.Rformat(self.bk_L1(x.get()),x.off,x.axis,chans=x.chans)
        else:
            if x.dtype==self.all_cbk_type:
                xr=self.bk_real(x)
                xi=self.bk_imag(x)
                
                r=self.backend.sign(xr)*self.backend.sqrt(self.backend.sign(xr)*xr)
                i=self.backend.sign(xi)*self.backend.sqrt(self.backend.sign(xi)*xi)
                return self.bk_complex(r,i)
            else:
                return self.backend.sign(x)*self.backend.sqrt(self.backend.sign(x)*x)
        
    def bk_reduce_sum(self,data,axis=None):
        
        if isinstance(data,Rformat.Rformat):
            return self.bk_reduce_sum(data.get())
        
        if axis is None:
            if self.BACKEND==self.TENSORFLOW:
                return(self.backend.reduce_sum(data))
            if self.BACKEND==self.TORCH:
                return(self.backend.sum(data))
            if self.BACKEND==self.NUMPY:
                return(np.sum(data))
        else:
            if self.BACKEND==self.TENSORFLOW:
                return(self.backend.reduce_sum(data,axis=axis))
            if self.BACKEND==self.TORCH:
                return(self.backend.sum(data,axis))
            if self.BACKEND==self.NUMPY:
                return(np.sum(data,axis))
        
    def bk_reduce_mean(self,data,axis=None):
        
        if isinstance(data,Rformat.Rformat):
            return self.bk_reduce_mean(data.get())
        
        if axis is None:
            if self.BACKEND==self.TENSORFLOW:
                return(self.backend.reduce_mean(data))
            if self.BACKEND==self.TORCH:
                return(self.backend.mean(data))
            if self.BACKEND==self.NUMPY:
                return(np.mean(data))
        else:
            if self.BACKEND==self.TENSORFLOW:
                return(self.backend.reduce_mean(data,axis=axis))
            if self.BACKEND==self.TORCH:
                return(self.backend.mean(data,axis))
            if self.BACKEND==self.NUMPY:
                return(np.mean(data,axis))
        
    def bk_sqrt(self,data):
        
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_sqrt(data.get()),data.off,0,chans=data.chans)
        
        return(self.backend.sqrt(self.backend.abs(data)))
    
    def bk_abs(self,data):
        
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_abs(data.get()),data.off,0,chans=data.chans)
        return(self.backend.abs(data))
        
    def bk_square(self,data):
        
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_square(data.get()),data.off,0,chans=data.chans)
        
        if self.BACKEND==self.TENSORFLOW:
            return(self.backend.square(data))
        if self.BACKEND==self.TORCH:
            return(self.backend.square(data))
        if self.BACKEND==self.NUMPY:
            return(data*data)
        
    def bk_log(self,data):
        if self.BACKEND==self.TENSORFLOW:
            return(self.backend.math.log(data))
        if self.BACKEND==self.TORCH:
            return(self.backend.log(data))
        if self.BACKEND==self.NUMPY:
            return(np.log(data))
        
    def bk_complex(self,real,imag):
        if self.BACKEND==self.TENSORFLOW:
            return(self.backend.dtypes.complex(real,imag))
        if self.BACKEND==self.TORCH:
            return(self.backend.complex(real,imag))
        if self.BACKEND==self.NUMPY:
            return(np.complex(real,imag))

    def bk_gather(self,data,shape,axis=None):
        if self.BACKEND==self.TENSORFLOW:
            if axis is None:
                return(self.backend.gather(data,shape))
            else:
                return(self.backend.gather(data,shape,axis=axis))
        if self.BACKEND==self.TORCH:
            my_tensor = self.backend.LongTensor(shape)
            my_data = self.backend.Tensor(data)
            
            return(self.backend.gather(my_data,axis,my_tensor))
        
        if self.BACKEND==self.NUMPY:
            if axis is None:
                return(np.take(data,shape))
            else:
                return(np.take(data,shape,axis=axis))

    def bk_reshape(self,data,shape):
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_reshape(data.get(),shape),data.off,0,chans=data.chans)
        
        return(self.backend.reshape(data,shape))
    
    def bk_repeat(self,data,nn,axis=0):
        return(self.backend.repeat(data,nn,axis=axis))
    
    def bk_roll(self,data,nn,axis=0):
        return(self.backend.roll(data,nn,axis=axis))

    def bk_expand_dims(self,data,axis=0,chans=0):
        if isinstance(data,Rformat.Rformat):
            if axis<data.axis:
                l_axis=data.axis+1
            else:
                l_axis=data.axis
            return Rformat.Rformat(self.bk_expand_dims(data.get(),axis=axis),data.off,l_axis,chans=data.chans)
            
        if self.BACKEND==self.TENSORFLOW:
            return(self.backend.expand_dims(data,axis=axis))
        if self.BACKEND==self.TORCH:
            return(self.backend.unsqueeze(data,axis))
        if self.BACKEND==self.NUMPY:
            return(np.expand_dims(data,axis))

    def bk_transpose(self,data,thelist):
        if self.BACKEND==self.TENSORFLOW:
            return(self.backend.transpose(data,thelist))
        if self.BACKEND==self.TORCH:
            return(self.backend.transpose(data,thelist))
        if self.BACKEND==self.NUMPY:
            return(np.transpose(data,thelist))

    def bk_concat(self,data,axis=None):
        if isinstance(data[0],Rformat.Rformat):
            l_data=[idata.get() for idata in data]
            return Rformat.Rformat(self.bk_concat(l_data,axis=axis),data[0].off,data[0].axis,chans=data[0].chans)
                
        if axis is None:
            return(self.backend.concat(data))
        else:
            return(self.backend.concat(data,axis=axis))

    
    def bk_conjugate(self,data):
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_conjugate(data.get()),data.off,data.axis,chans=data.chans)
                
        if self.BACKEND==self.TENSORFLOW:
            return self.backend.math.conj(data)
        if self.BACKEND==self.TORCH:
            return self.backend.conjugate(data)
        if self.BACKEND==self.NUMPY:
            return data.conjugate()
        
    def bk_real(self,data):
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_real(data.get()),data.off,data.axis,chans=data.chans)
                
        if self.BACKEND==self.TENSORFLOW:
            return self.backend.math.real(data)
        if self.BACKEND==self.TORCH:
            return self.backend.real(data)
        if self.BACKEND==self.NUMPY:
            return self.backend.real(data)

    def bk_imag(self,data):
        if isinstance(data,Rformat.Rformat):
            return Rformat.Rformat(self.bk_imag(data.get()),data.off,data.axis,chans=data.chans)
                
        if self.BACKEND==self.TENSORFLOW:
            return self.backend.math.imag(data)
        if self.BACKEND==self.TORCH:
            return self.backend.imag(data)
        if self.BACKEND==self.NUMPY:
            return self.backend.imag(data)
        
    def bk_relu(self,x):
        if isinstance(x,Rformat.Rformat):
            return Rformat.Rformat(self.bk_relu(x.get()),x.off,x.axis,chans=x.chans)
        
        if self.BACKEND==self.TENSORFLOW:
            if x.dtype==self.all_cbk_type:
                xr=self.backend.nn.relu(self.bk_real(x))
                xi=self.backend.nn.relu(self.bk_imag(x))
                return self.backend.complex(xr,xi)
            else:
                return self.backend.nn.relu(x)
        if self.BACKEND==self.TORCH:
            return self.backend.relu(x)
        if self.BACKEND==self.NUMPY:
            return (x>0)*x
        
    def bk_cast(self,x):
        if isinstance(x,np.float64):
            if self.all_bk_type=='float32':
                return(np.float32(x))
            else:
                return(x)
        if isinstance(x,np.float32):
            if self.all_bk_type=='float64':
                return(np.float64(x))
            else:
                return(x)
        
        if isinstance(x,Rformat.Rformat):
            return Rformat.Rformat(self.bk_cast(x.get()),x.off,x.axis,chans=x.chans)

        if x.dtype=='complex128' or x.dtype=='complex64':
            out_type=self.all_cbk_type
        else:
            out_type=self.all_bk_type
            
        if self.BACKEND==self.TENSORFLOW:
            return self.backend.cast(x,out_type)
        if self.BACKEND==self.TORCH:
            return self.backend.cast(x,out_type)
        if self.BACKEND==self.NUMPY:
            return x.astype(out_type)



import foscat.FoCUS as FOC
import numpy as np
import foscat.backend as bk
import foscat.Rformat as Rformat


def read(filename):
    thescat = scat_cov(1, 1, 1)
    return thescat.read(filename)


class scat_cov:
    def __init__(self, p00, c01, c11, s1=None, c10=None,backend=None):
        self.P00     = p00
        self.C01     = c01
        self.C11     = c11
        self.S1      = s1
        self.C10     = c10
        self.backend = backend

    def numpy(self):
        if self.S1 is None:
            s1 = None
        else:
            s1=self.S1.numpy()
        if self.C10 is None:
            c10 = None
        else:
            c10=self.C10.numpy()
        
        return scat_cov((self.P00.numpy()),
                        (self.C01.numpy()),
                        (self.C11.numpy()),
                        s1=s1, c10=c10,backend=self.backend)
        
    def get_S1(self):
        return self.S1

    def get_P00(self):
        return self.P00

    def reset_P00(self):
        self.P00=0*self.P00

    def get_C01(self):
        return self.C01

    def get_C10(self):
        return self.C10

    def get_C11(self):
        return self.C11

    def __add__(self, other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
               isinstance(other, bool) or isinstance(other, scat_cov)

        if self.S1 is None:
            s1 = None
        else:
            if isinstance(other, scat_cov):
                if other.S1 is None:
                    s1=None
                else:
                    s1 = self.S1 + other.S1
            else:
                s1 = self.S1 + other

        if self.C10 is None:
            c10 = None
        else:
            if isinstance(other, scat_cov):
                if other.C10 is None:
                    c10=None
                else:
                    c10 = self.C10 + other.C10
            else:
                c10 = self.C10 + other

        if isinstance(other, scat_cov):
            return scat_cov(self.doadd(self.P00,other.P00),
                            (self.C01 + other.C01),
                            (self.C11 + other.C11),
                            s1=s1, c10=c10,backend=self.backend)
        else:
            return scat_cov((self.P00 + other),
                            (self.C01 + other),
                            (self.C11 + other),
                            s1=s1, c10=c10,backend=self.backend)

    def __radd__(self, other):
        return self.__add__(other)
    
    def __truediv__(self, other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
               isinstance(other, bool) or isinstance(other, scat_cov)

        if self.S1 is None:
            s1 = None
        else:
            if isinstance(other, scat_cov):
                if other.S1 is None:
                    s1 = None
                else:
                    s1 = self.S1 / other.S1
            else:
                s1 = self.S1 / other

        if self.C10 is None:
            c10 = None
        else:
            if isinstance(other, scat_cov):
                if other.C10 is None:
                    c10 = None
                else:
                    c10 = self.C10 / other.C10
            else:
                c10 = self.C10 / other

        if isinstance(other, scat_cov):
            return scat_cov(self.dodiv(self.P00,other.P00),
                            (self.C01 / other.C01),
                            (self.C11 / other.C11),
                            s1=s1, c10=c10,backend=self.backend)
        else:
            return scat_cov((self.P00 / other),
                            (self.C01 / other),
                            (self.C11 / other),
                            s1=s1, c10=c10,backend=self.backend)

    def __rtruediv__(self, other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
               isinstance(other, bool) or isinstance(other, scat_cov)

        if self.S1 is None:
            s1 = None
        else:
            if isinstance(other, scat_cov):
                s1 = other.S1 / self.S1
            else:
                s1 = other.S1 / other

        if self.C10 is None:
            c10 = None
        else:
            if isinstance(other, scat_cov):
                c10 = other.C10 / self.C10
            else:
                c10 = other/self.C10

        if isinstance(other, scat_cov):
            return scat_cov(self.dodiv(other.P00,self.P00),
                            (other.C01 / self.C01),
                            (other.C11 / self.C11),
                            s1=s1, c10=c10,backend=self.backend)
        else:
            return scat_cov((other/self.P00 ),
                            (other/self.C01 ),
                            (other/self.C11 ),
                            s1=s1, c10=c10,backend=self.backend)

    def __rsub__(self, other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
               isinstance(other, bool) or isinstance(other, scat_cov)

        if self.S1 is None:
            s1 = None
        else:
            if isinstance(other, scat_cov):
                if other.S1 is None:
                    s1 = None
                else:
                    s1 = other.S1 - self.S1
            else:
                s1 = other - self.S1

        if self.C10 is None:
            c10 = None
        else:
            if isinstance(other, scat_cov):
                if other.C10 is None:
                    c10 = None
                else:
                    c10 = other.C10 - self.C10 
            else:
                c10 = other - self.C10

        if isinstance(other, scat_cov):
            return scat_cov(self.domin(other.P00,self.P00),
                            (other.C01 - self.C01),
                            (other.C11 - self.C11),
                            s1=s1, c10=c10,
                            backend=self.backend)
        else:
            return scat_cov((other-self.P00),
                            (other-self.C01),
                            (other-self.C11),
                            s1=s1, c10=c10,
                            backend=self.backend)
        
    def __sub__(self, other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
               isinstance(other, bool) or isinstance(other, scat_cov)

        if self.S1 is None:
            s1 = None
        else:
            if isinstance(other, scat_cov):
                if other.S1 is None:
                    s1 = None
                else:
                    s1 = self.S1 - other.S1
            else:
                s1 = self.S1 - other

        if self.C10 is None:
            c10 = None
        else:
            if isinstance(other, scat_cov):
                if other.C10 is None:
                    c10 = None
                else:
                    c10 = self.C10 - other.C10
            else:
                c10 = self.C10 - other

        if isinstance(other, scat_cov):
            return scat_cov(self.domin(self.P00,other.P00),
                            (self.C01 - other.C01),
                            (self.C11 - other.C11),
                            s1=s1, c10=c10,backend=self.backend)
        else:
            return scat_cov((self.P00 - other),
                            (self.C01 - other),
                            (self.C11 - other),
                            s1=s1, c10=c10,backend=self.backend)

    def domult(self,x,y):
        if x.dtype==y.dtype:
            return x*y
        if x.dtype=='complex64' or x.dtype=='complex128':
            
            return self.backend.bk_complex(self.backend.bk_real(x)*y,self.backend.bk_imag(x)*y)
        else:
            return self.backend.bk_complex(self.backend.bk_real(y)*x,self.backend.bk_imag(y)*x)
        
    def dodiv(self,x,y):
        if x.dtype==y.dtype:
            return x/y
        if x.dtype=='complex64' or x.dtype=='complex128':
            
            return self.backend.bk_complex(self.backend.bk_real(x)/y,self.backend.bk_imag(x)/y)
        else:
            return self.backend.bk_complex(x/self.backend.bk_real(y),x/self.backend.bk_imag(y))
        
    def domin(self,x,y):
        if x.dtype==y.dtype:
            return x-y
        if x.dtype=='complex64' or x.dtype=='complex128':
            
            return self.backend.bk_complex(self.backend.bk_real(x)-y,self.backend.bk_imag(x)-y)
        else:
            return self.backend.bk_complex(x-self.backend.bk_real(y),x-self.backend.bk_imag(y))
        
    def doadd(self,x,y):
        if x.dtype==y.dtype:
            return x+y
        if x.dtype=='complex64' or x.dtype=='complex128':
            
            return self.backend.bk_complex(self.backend.bk_real(x)+y,self.backend.bk_imag(x)+y)
        else:
            return self.backend.bk_complex(x+self.backend.bk_real(y),x+self.backend.bk_imag(y))
                
            
    def __mul__(self, other):
        assert isinstance(other, float)  or isinstance(other, np.float32) or isinstance(other, int) or \
               isinstance(other, bool) or isinstance(other, scat_cov)

        if self.S1 is None:
            s1 = None
        else:
            if isinstance(other, scat_cov):
                if other.S1 is None:
                    s1 = None
                else:
                    s1 = self.S1 * other.S1
            else:
                s1 = self.S1 * other

        if self.C10 is None:
            c10 = None
        else:
            if isinstance(other, scat_cov):
                if other.C10 is None:
                    c10 = None
                else:
                    c10 = self.C10 * other.C10
            else:
                c10 = self.C10 * other

        if isinstance(other, scat_cov):
            return scat_cov(self.domult(self.P00,other.P00),
                            (self.C01 * other.C01),
                            (self.C11 * other.C11),
                            s1=s1, c10=c10,backend=self.backend)
        else:
            return scat_cov((self.P00 * other),
                            (self.C01 * other),
                            (self.C11 * other),
                            s1=s1, c10=c10,backend=self.backend)

    
    def __rmul__(self, other):
        return self.__mul__(other)

    def plot(self, name=None, hold=True, color='blue', lw=1, legend=True):

        import matplotlib.pyplot as plt

        if name is None:
            name = ''

        if hold:
            plt.figure(figsize=(8, 8))

        if self.S1 is not None:
            plt.subplot(2, 2, 1)
            if legend:
                plt.plot(abs(self.get_np(self.S1)).flatten(), color=color, label=r'%s $S_1$' % (name), lw=lw)
            else:
                plt.plot(abs(self.get_np(self.S1)).flatten(), color=color, lw=lw)
            plt.yscale('log')
            plt.legend()

        plt.subplot(2, 2, 2)
        if legend:
            plt.plot(abs(self.get_np(self.P00)).flatten(), color=color, label=r'%s $P_{00}$' % (name), lw=lw)
        else:
            plt.plot(abs(self.get_np(self.P00)).flatten(), color=color, lw=lw)
        plt.yscale('log')
        plt.legend()

        plt.subplot(2, 2, 3)
        if legend:
            plt.plot(abs(self.get_np(self.C01)).flatten(), color=color, label=r'%s $C_{01}$' % (name), lw=lw)
            if self.C10 is not None:
                plt.plot(abs(self.get_np(self.C10)).flatten(), color=color, label=r'%s $C_{10}$' % (name), lw=lw)
        else:
            plt.plot(abs(self.get_np(self.C01)).flatten(), color=color, lw=lw)
            if self.C10 is not None:
                plt.plot(abs(self.get_np(self.C10)).flatten(), color=color, lw=lw)
        plt.yscale('log')
        plt.legend()

        plt.subplot(2, 2, 4)
        if legend:
            plt.plot(abs(self.get_np(self.C11)).flatten(), color=color, label=r'%s $C_{11}$' % (name), lw=lw)
        else:
            plt.plot(abs(self.get_np(self.C11)).flatten(), color=color, lw=lw)
        plt.yscale('log')
        plt.legend()

    def get_np(self, x):
        if isinstance(x, np.ndarray):
            return x
        else:
            return x.numpy()
        
    def save(self, filename):
        if self.S1 is not None:  # Auto
            np.save('%s_s1.npy' % filename, self.get_np(self.get_S1()))
        if self.C10 is not None:  # Cross
            np.save('%s_c10.npy' % filename, self.get_np(self.get_C10()))
        np.save('%s_c01.npy' % filename, self.get_np(self.get_C01()))
        np.save('%s_c11.npy' % filename, self.get_np(self.get_C11()))
        np.save('%s_p0.npy' % filename, self.get_np(self.get_P00()))

    def read(self, filename):
        try:
            s1 = np.load('%s_s1.npy' % filename)
        except:
            s1 = None

        try:
            c10 = np.load('%s_c10.npy' % filename)
        except:
            c10 = None

        c01 = np.load('%s_c01.npy' % filename)
        c11 = np.load('%s_c11.npy' % filename)
        p0 = np.load('%s_p0.npy' % filename)

        return scat_cov(p0, c01, c11, s1=s1, c10=c10,backend=self.backend)

    def std(self):
        if self.S1 is not None:  # Auto
            return np.sqrt(((abs(self.get_np(self.S1)).std()) ** 2 +
                            (abs(self.get_np(self.C01)).std()) ** 2 +
                            (abs(self.get_np(self.C11)).std()) ** 2 +
                            (abs(self.get_np(self.P00)).std()) ** 2 ) / 4)
        else:  # Cross
            return np.sqrt(((abs(self.get_np(self.C01)).std()) ** 2 +
                            (abs(self.get_np(self.C10)).std()) ** 2 +
                            (abs(self.get_np(self.C11)).std()) ** 2 +
                            (abs(self.get_np(self.P00)).std()) ** 2) / 4)

    def mean(self):
        if self.S1 is not None:  # Auto
            return (abs(self.get_np(self.S1)).mean() +
                    abs(self.get_np(self.C01)).mean() +
                    abs(self.get_np(self.C11)).mean() +
                    abs(self.get_np(self.P00)).mean()) / 4
        else:  # Cross
            return (abs(self.get_np(self.C01)).mean() +
                    abs(self.get_np(self.C10)).mean() +
                    abs(self.get_np(self.C11)).mean() +
                    abs(self.get_np(self.P00)).mean()) / 4

    def get_nscale(self):
        return self.P00.shape[2]
    
    def get_norient(self):
        return self.P00.shape[3]

    def add_data_from_log_slope(self,y,n,ds=3):
        if len(y)<ds:
            if len(y)==1:
                return(np.repeat(y[0],n))
            if len(y)==2:
                a=np.polyfit(np.arange(2),np.log(y[0:2]),1)
        else:
            a=np.polyfit(np.arange(ds),np.log(y[0:ds]),1)
        return np.exp((np.arange(n)-1-n)*a[0]+a[1])

    def add_data_from_slope(self,y,n,ds=3):
        if len(y)<ds:
            if len(y)==1:
                return(np.repeat(y[0],n))
            if len(y)==2:
                a=np.polyfit(np.arange(2),y[0:2],1)
        else:
            a=np.polyfit(np.arange(ds),y[0:ds],1)
        return (np.arange(n)-1-n)*a[0]+a[1]
    
    def up_grade(self,nscale,ds=3):
        noff=nscale-self.P00.shape[2]
        if noff==0:
            return scat_cov((self.P00),
                            (self.C01),
                            (self.C11),
                            s1=self.S1,
                            c10=self.C10,backend=self.backend)
        
        inscale=self.P00.shape[2]
        p00=np.zeros([self.P00.shape[0],self.P00.shape[1],nscale,self.P00.shape[3]],dtype='complex')
        p00[:,:,noff:,:]=self.P00.numpy()
        for i in range(self.P00.shape[0]):
            for j in range(self.P00.shape[1]):
                for k in range(self.P00.shape[3]):
                    p00[i,j,0:noff,k]=self.add_data_from_log_slope(p00[i,j,noff:,k],noff,ds=ds)
                    
        s1=np.zeros([self.S1.shape[0],self.S1.shape[1],nscale,self.S1.shape[3]])
        s1[:,:,noff:,:]=self.S1.numpy()
        for i in range(self.S1.shape[0]):
            for j in range(self.S1.shape[1]):
                for k in range(self.S1.shape[3]):
                    s1[i,j,0:noff,k]=self.add_data_from_log_slope(s1[i,j,noff:,k],noff,ds=ds)

        nout=0
        for i in range(1,nscale):
            nout=nout+i
            
        c01=np.zeros([self.C01.shape[0],self.C01.shape[1], \
                      nout,self.C01.shape[3],self.C01.shape[4]],dtype='complex')
                     
        jo1=np.zeros([nout])
        jo2=np.zeros([nout])

        n=0
        for i in range(1,nscale):
            jo1[n:n+i]=np.arange(i)
            jo2[n:n+i]=i
            n=n+i
            
        j1=np.zeros([self.C01.shape[2]])
        j2=np.zeros([self.C01.shape[2]])
        
        n=0
        for i in range(1,self.P00.shape[2]):
            j1[n:n+i]=np.arange(i)
            j2[n:n+i]=i
            n=n+i

        for i in range(self.C01.shape[0]):
            for j in range(self.C01.shape[1]):
                for k in range(self.C01.shape[3]):
                    for l in range(self.C01.shape[4]):
                        for ij in range(noff+1,nscale):
                            idx=np.where(jo2==ij)[0]
                            c01[i,j,idx[noff:],k,l]=self.C01.numpy()[i,j,j2==ij-noff,k,l]
                            c01[i,j,idx[:noff],k,l]=self.add_data_from_slope(self.C01.numpy()[i,j,j2==ij-noff,k,l],noff,ds=ds)

                        for ij in range(nscale):
                            idx=np.where(jo1==ij)[0]
                            if idx.shape[0]>noff:
                                c01[i,j,idx[:noff],k,l]=self.add_data_from_slope(c01[i,j,idx[noff:],k,l],noff,ds=ds)
                            else:
                                c01[i,j,idx,k,l]=np.mean(c01[i,j,jo1==ij-1,k,l])

        
        nout=0
        for j3 in range(nscale):
            for j2 in range(0,j3):
                for j1 in range(0,j2):
                    nout=nout+1

        c11=np.zeros([self.C11.shape[0],self.C11.shape[1], \
                      nout,self.C11.shape[3], \
                      self.C11.shape[4],self.C11.shape[5]],dtype='complex')
                     
        jo1=np.zeros([nout])
        jo2=np.zeros([nout])
        jo3=np.zeros([nout])

        nout=0
        for j3 in range(nscale):
            for j2 in range(0,j3):
                for j1 in range(0,j2):
                    jo1[nout]=j1
                    jo2[nout]=j2
                    jo3[nout]=j3
                    nout=nout+1

        ncross=self.C11.shape[2]
        jj1=np.zeros([ncross])
        jj2=np.zeros([ncross])
        jj3=np.zeros([ncross])
        
        n=0
        for j3 in range(inscale):
            for j2 in range(0,j3):
                for j1 in range(0,j2):
                    jj1[n]=j1
                    jj2[n]=j2
                    jj3[n]=j3
                    n=n+1

        n=0
        for j3 in range(nscale):
            for j2 in range(j3):
                idx=np.where((jj3==j3)*(jj2==j2))[0]
                if idx.shape[0]>0:
                    idx2=np.where((jo3==j3+noff)*(jo2==j2+noff))[0]
                    for i in range(self.C11.shape[0]):
                        for j in range(self.C11.shape[1]):
                            for k in range(self.C11.shape[3]):
                                for l in range(self.C11.shape[4]):
                                    for m in range(self.C11.shape[5]):
                                        c11[i,j,idx2[noff:],k,l,m]=self.C11.numpy()[i,j,idx,k,l,m]
                                        c11[i,j,idx2[:noff],k,l,m]=self.add_data_from_log_slope(self.C11.numpy()[i,j,idx,k,l,m],noff,ds=ds)

        idx=np.where(abs(c11[0,0,:,0,0,0])==0)[0]
        for iii in idx:
            iii1=np.where((jo1==jo1[iii]+1)*(jo2==jo2[iii]+1)*(jo3==jo3[iii]+1))[0]
            iii2=np.where((jo1==jo1[iii]+2)*(jo2==jo2[iii]+2)*(jo3==jo3[iii]+2))[0]
            if iii2.shape[0]>0:
                for i in range(self.C11.shape[0]):
                    for j in range(self.C11.shape[1]):
                        for k in range(self.C11.shape[3]):
                            for l in range(self.C11.shape[4]):
                                for m in range(self.C11.shape[5]):
                                    c11[i,j,iii,k,l,m]=self.add_data_from_slope(c11[i,j,[iii1,iii2],k,l,m],1,ds=2)[0]
        
        idx=np.where(abs(c11[0,0,:,0,0,0])==0)[0]
        for iii in idx:
            iii1=np.where((jo1==jo1[iii])*(jo2==jo2[iii])*(jo3==jo3[iii]-1))[0]
            iii2=np.where((jo1==jo1[iii])*(jo2==jo2[iii])*(jo3==jo3[iii]-2))[0]
            if iii2.shape[0]>0:
                for i in range(self.C11.shape[0]):
                    for j in range(self.C11.shape[1]):
                        for k in range(self.C11.shape[3]):
                            for l in range(self.C11.shape[4]):
                                for m in range(self.C11.shape[5]):
                                    c11[i,j,iii,k,l,m]=self.add_data_from_slope(c11[i,j,[iii1,iii2],k,l,m],1,ds=2)[0]

        return scat_cov( (p00),
                         (c01),
                         (c11),
                        s1=(s1),backend=self.backend)
        
        

class funct(FOC.FoCUS):

    def eval(self, image1, image2=None, mask=None, norm=None, Auto=True):
        """
        Calculates the scattering correlations for a batch of images. Mean are done over pixels.
        mean of modulus:
                        S1 = <|I * Psi_j3|>
             Normalization : take the log
        power spectrum:
                        P00 = <|I * Psi_j3|^2>
            Normalization : take the log
        orig. x modulus:
                        C01 = < (I * Psi)_j3 x (|I * Psi_j2| * Psi_j3)^* >
             Normalization : divide by (P00_j2 * P00_j3)^0.5
        modulus x modulus:
                        C11 = <(|I * psi1| * psi3)(|I * psi2| * psi3)^*>
             Normalization : divide by (P00_j1 * P00_j2)^0.5
        Parameters
        ----------
        image1: tensor
            Image on which we compute the scattering coefficients [Nbatch, Npix, 1, 1]
        image2: tensor
            Second image. If not None, we compute cross-scattering covariance coefficients.
        mask:
        norm: None or str
            If None no normalization is applied, if 'auto' normalize by the reference P00,
            if 'self' normalize by the current P00.
        all_cross: False or True
            If False compute all the coefficient even the Imaginary part,
            If True return only the terms computable in the auto case.
        Returns
        -------
        S1, P00, C01, C11 normalized
        """
        ### AUTO OR CROSS
        cross = False
        if image2 is not None:
            cross = True
            all_cross=Auto
        else:
            all_cross=False

        ### PARAMETERS
        axis = 1
        # determine jmax and nside corresponding to the input map
        im_shape = image1.shape
        if self.use_R_format and isinstance(image1, Rformat.Rformat):
            if len(image1.shape) == 4:
                nside = im_shape[2] - 2 * self.R_off
            else:
                nside = im_shape[1] - 2 * self.R_off
            npix = 12 * nside * nside  # Number of pixels
        else:
            if self.chans==1:
                nside=im_shape[axis]
                npix=nside*nside
            else:
                npix = image1.shape[-1]  # image1 is [Npix] or [Nbatch, Npix]
                nside = int(np.sqrt(npix / 12))

        J = int(np.log(nside) / np.log(2))  # Number of j scales
        Jmax = J - self.OSTEP  # Number of steps for the loop on scales

        ### LOCAL VARIABLES (IMAGES and MASK)
        # Check if image1 is [Npix] or [Nbatch, Npix] or Rformat
        if len(image1.shape) == 1 or (len(image1.shape)==2 and self.chans==1) or (len(image1.shape) == 3 and isinstance(image1, Rformat.Rformat)):
            I1 = self.backend.bk_cast(self.backend.bk_expand_dims(image1, 0))  # Local image1 [Nbatch, Npix]
            if cross:
                I2 = self.backend.bk_cast(self.backend.bk_expand_dims(image2, 0))  # Local image2 [Nbatch, Npix]
        else:
            I1 = self.backend.bk_cast(image1)  # Local image1 [Nbatch, Npix]
            if cross:
                I2 = self.backend.bk_cast(image2)  # Local image2 [Nbatch, Npix]
                
        if mask is None:
            if self.chans==1:
                vmask = self.backend.bk_ones([1, nside, nside],dtype=self.all_type)
            else:
                vmask = self.backend.bk_ones([1, npix], dtype=self.all_type)
                
            if self.use_R_format:
                vmask = self.to_R(vmask, axis=1,chans=self.chans)
        else:
            vmask = self.backend.bk_cast(mask)  # [Nmask, Npix]
            if self.use_R_format:
                vmask = self.to_R(vmask, axis=1,chans=self.chans)

        if self.use_R_format:
            I1 = self.to_R(I1, axis=axis,chans=self.chans)
            if cross:
                I2 = self.to_R(I2, axis=axis,chans=self.chans)

        if self.KERNELSZ > 3:
            # if the kernel size is bigger than 3 increase the binning before smoothing
            I1 = self.up_grade(I1, nside * 2, axis=axis)
            vmask = self.up_grade(vmask, nside * 2, axis=1)
            if cross:
                I2 = self.up_grade(I2, nside * 2, axis=axis)
                
        # Normalize the masks because they have different pixel numbers
        # vmask /= self.backend.bk_reduce_sum(vmask, axis=1)[:, None]  # [Nmask, Npix]

        ### INITIALIZATION
        # Coefficients
        S1, P00, C01, C11, C10 = None, None, None, None, None

        # Dictionaries for C01 computation
        M1_dic = {}  # M stands for Module M1 = |I1 * Psi|
        if cross:
            M2_dic = {}

        # P00 for normalization
        cond_init_P1_dic = ((norm == 'self') or ((norm == 'auto') and (self.P1_dic is None)))
        if norm is None:
            pass
        elif cond_init_P1_dic:
            P1_dic = {}
            if cross:
                P2_dic = {}
        elif (norm == 'auto') and (self.P1_dic is not None):
            P1_dic = self.P1_dic
            if cross:
                P2_dic = self.P2_dic

        #### COMPUTE S1, P00, C01 and C11
        nside_j3 = nside  # NSIDE start (nside_j3 = nside / 2^j3)
        for j3 in range(Jmax):

            ####### S1 and P00
            ### Make the convolution I1 * Psi_j3
            conv1 = self.convol(I1, axis=1)  # [Nbatch, Npix_j3, Norient3]
            ### Take the module M1 = |I1 * Psi_j3|
            M1_square = conv1*self.backend.bk_conjugate(conv1) # [Nbatch, Npix_j3, Norient3]
            M1 = self.backend.bk_L1(M1_square)  # [Nbatch, Npix_j3, Norient3]
            # Store M1_j3 in a dictionary
            M1_dic[j3] = self.update_R_border(M1, axis=axis)

            if not cross:  # Auto
                M1_square=self.backend.bk_real(M1_square)
                
                ### P00_auto = < M1^2 >_pix
                # Apply the mask [Nmask, Npix_j3] and average over pixels
                p00 = self.masked_mean(M1_square, vmask, axis=1)
                if cond_init_P1_dic:
                    # We fill P1_dic with P00 for normalisation of C01 and C11
                    P1_dic[j3] = p00  # [Nbatch, Nmask, Norient3]
                if norm == 'auto':  # Normalize P00
                    p00 /= P1_dic[j3]

                # We store P00_auto to return it [Nbatch, Nmask, NP00, Norient3]
                if P00 is None:
                    P00 = p00[:, :, None, :]  # Add a dimension for NP00
                else:
                    P00 = self.backend.bk_concat([P00, p00[:, :, None, :]], axis=2)

                #### S1_auto computation
                ### Image 1 : S1 = < M1 >_pix
                # Apply the mask [Nmask, Npix_j3] and average over pixels
                s1 = self.masked_mean(M1, vmask, axis=1)  # [Nbatch, Nmask, Norient3]
                ### Normalize S1
                if norm is not None:
                    s1 /= (P1_dic[j3]) ** 0.5
                ### We store S1 for image1  [Nbatch, Nmask, NS1, Norient3]
                if S1 is None:
                    S1 = s1[:, :, None, :]  # Add a dimension for NS1
                else:
                    S1 = self.backend.bk_concat([S1, s1[:, :, None, :]], axis=2)

            else:  # Cross
                ### Make the convolution I2 * Psi_j3
                conv2 = self.convol(I2, axis=1)  # [Nbatch, Npix_j3, Norient3]
                ### Take the module M2 = |I2 * Psi_j3|
                M2_square = conv2*self.backend.bk_conjugate(conv2)  # [Nbatch, Npix_j3, Norient3]
                M2 = self.backend.bk_L1(M2_square)  # [Nbatch, Npix_j3, Norient3]
                # Store M2_j3 in a dictionary
                M2_dic[j3] = self.update_R_border(M2, axis=axis)

                ### P00_auto = < M2^2 >_pix
                # Not returned, only for normalization
                if cond_init_P1_dic:
                    # Apply the mask [Nmask, Npix_j3] and average over pixels
                    p1 = self.masked_mean(M1_square, vmask, axis=1)  # [Nbatch, Nmask, Norient3]
                    p2 = self.masked_mean(M2_square, vmask, axis=1)  # [Nbatch, Nmask, Norient3]
                    # We fill P1_dic with P00 for normalisation of C01 and C11
                    P1_dic[j3] = p1  # [Nbatch, Nmask, Norient3]
                    P2_dic[j3] = p2  # [Nbatch, Nmask, Norient3]

                ### P00_cross = < (I1 * Psi_j3) (I2 * Psi_j3)^* >_pix
                # z_1 x z_2^* = (a1a2 + b1b2) + i(b1a2 - a1b2)
                p00 = conv1 * self.backend.bk_conjugate(conv2)
                # Apply the mask [Nmask, Npix_j3] and average over pixels
                p00 = self.masked_mean(p00, vmask, axis=1)

                ### Normalize P00_cross
                if norm == 'auto':
                    p00 /= (P1_dic[j3] * P2_dic[j3])**0.5

                ### Store P00_cross as complex [Nbatch, Nmask, NP00, Norient3]
                if not all_cross:
                    p00=self.backend.bk_real(p00)
                    
                if P00 is None:
                    P00 = p00[:,:,None,:]  # Add a dimension for NP00
                else:
                    P00 = self.backend.bk_concat([P00, p00[:,:,None,:]], axis=2)
                    
            # Initialize dictionaries for |I1*Psi_j| * Psi_j3
            M1convPsi_dic = {}
            if cross:
                # Initialize dictionaries for |I2*Psi_j| * Psi_j3
                M2convPsi_dic = {}

            ###### C01
            for j2 in range(0, j3):  # j2 < j3
                ### C01_auto = < (I1 * Psi)_j3 x (|I1 * Psi_j2| * Psi_j3)^* >_pix
                if not cross:
                    c01 = self._compute_C01(j2,
                                            conv1,
                                            vmask,
                                            M1_dic,
                                            M1convPsi_dic)  # [Nbatch, Nmask, Norient3, Norient2]
                    ### Normalize C01 with P00_j [Nbatch, Nmask, Norient_j]
                    if norm is not None:
                        c01 /= (P1_dic[j2][:, :, None, :] *
                                P1_dic[j3][:, :, :, None]) ** 0.5  # [Nbatch, Nmask, Norient3, Norient2]
                        
                    ### Store C01 as a complex [Nbatch, Nmask, NC01, Norient3, Norient2]
                    if C01 is None:
                        C01 = c01[:,:,None,:,:]  # Add a dimension for NC01
                    else:
                        C01 = self.backend.bk_concat([C01, c01[:, :, None, :, :]],
                                                 axis=2)  # Add a dimension for NC01

                ### C01_cross = < (I1 * Psi)_j3 x (|I2 * Psi_j2| * Psi_j3)^* >_pix
                ### C10_cross = < (I2 * Psi)_j3 x (|I1 * Psi_j2| * Psi_j3)^* >_pix
                else:
                    c01 = self._compute_C01(j2,
                                            conv1,
                                            vmask,
                                            M2_dic,
                                            M2convPsi_dic)
                    c10 = self._compute_C01(j2,
                                            conv2,
                                            vmask,
                                            M1_dic,
                                            M1convPsi_dic)
                    
                    ### Normalize C01 and C10 with P00_j [Nbatch, Nmask, Norient_j]
                    if norm is not None:
                        c01 /= (P2_dic[j2][:, :, None, :] *
                                P1_dic[j3][:, :, :, None]) ** 0.5  # [Nbatch, Nmask, Norient3, Norient2]
                        c10 /= (P1_dic[j2][:, :, None, :] *
                                P2_dic[j3][:, :, :, None]) ** 0.5  # [Nbatch, Nmask, Norient3, Norient2]
                        
                    ### Store C01 and C10 as a complex [Nbatch, Nmask, NC01, Norient3, Norient2]
                    if C01 is None:
                        C01 = c01[:, :, None, :, :] # Add a dimension for NC01
                    else:
                        C01 = self.backend.bk_concat([C01,c01[:, :, None, :, :]],axis=2)  # Add a dimension for NC01
                    if C10 is None:
                        C10 = c10[:, :, None, :, :]  # Add a dimension for NC01
                    else:
                        C10 = self.backend.bk_concat([C10,c10[:, :, None, :, :]], axis=2)  # Add a dimension for NC01
                        


                ##### C11
                for j1 in range(0, j2):  # j1 < j2
                    ### C11_auto = <(|I1 * psi1| * psi3)(|I1 * psi2| * psi3)^*>
                    if not cross:
                        c11 = self._compute_C11(j1, j2, vmask,
                                                M1convPsi_dic,
                                                M2convPsi_dic=None) # [Nbatch, Nmask, Norient3, Norient2, Norient1]
                        ### Normalize C11 with P00_j [Nbatch, Nmask, Norient_j]
                        if norm is not None:
                            c11 /= (P1_dic[j1][:, :, None, None, :] *
                                    P1_dic[j2][:, :, None, :,
                                               None]) ** 0.5  # [Nbatch, Nmask, Norient3, Norient2, Norient1]
                        ### Store C11 as a complex [Nbatch, Nmask, NC11, Norient3, Norient2, Norient1]
                        if C11 is None:
                            C11 = c11[:, :, None, :, :, :]  # Add a dimension for NC11
                        else:
                            C11 = self.backend.bk_concat([C11,c11[:, :, None, :, :, :]],
                                                 axis=2)  # Add a dimension for NC11

                        ### C11_cross = <(|I1 * psi1| * psi3)(|I2 * psi2| * psi3)^*>
                    else:
                        c11 = self._compute_C11(j1, j2, vmask,
                                                M1convPsi_dic,
                                                M2convPsi_dic=M2convPsi_dic)  # [Nbatch, Nmask, Norient3, Norient2, Norient1]
                        ### Normalize C11 with P00_j [Nbatch, Nmask, Norient_j]
                        if norm is not None:
                            c11 /= (P1_dic[j1][:, :, None, None, :] *
                                    P2_dic[j2][:, :, None, :, None]) ** 0.5  # [Nbatch, Nmask, Norient3, Norient2, Norient1]
                        ### Store C11 as a complex [Nbatch, Nmask, NC11, Norient3, Norient2, Norient1]
                        if C11 is None:
                            C11 = c11[:, :, None, :, :, :]  # Add a dimension for NC11
                        else:
                            C11 = self.backend.bk_concat([C11,c11[:, :, None, :, :, :]],
                                                 axis=2)  # Add a dimension for NC11
            
            ###### Reshape for next iteration on j3
            ### Image I1,
            # downscale the I1 [Nbatch, Npix_j3]
            if j3 != Jmax - 1:
                I1_smooth = self.smooth(I1, axis=1)
                I1 = self.ud_grade_2(I1_smooth, axis=1)

                ### Image I2
                if cross:
                    I2_smooth = self.smooth(I2, axis=1)
                    I2 = self.ud_grade_2(I2_smooth, axis=1)

                ### Modules
                for j2 in range(0, j3 + 1):  # j2 =< j3
                    ### Dictionary M1_dic[j2]
                    M1_smooth = self.smooth(M1_dic[j2], axis=1)  # [Nbatch, Npix_j3, Norient3]
                    M1_dic[j2] = self.ud_grade_2(M1_smooth, axis=1)  # [Nbatch, Npix_j3, Norient3]

                    ### Dictionary M2_dic[j2]
                    if cross:
                        M2_smooth = self.smooth(M2_dic[j2], axis=1)  # [Nbatch, Npix_j3, Norient3]
                        M2_dic[j2] = self.ud_grade_2(M2_smooth, axis=1)  # [Nbatch, Npix_j3, Norient3]
                ### Mask
                vmask = self.ud_grade_2(vmask, axis=1)

                ### NSIDE_j3
                nside_j3 = nside_j3 // 2

        ### Store P1_dic and P2_dic in self
        if (norm == 'auto') and (self.P1_dic is None):
            self.P1_dic = P1_dic
            if cross:
                self.P2_dic = P2_dic

        if not cross:
            return scat_cov(P00, C01, C11, s1=S1,backend=self.backend)
        else:
            return scat_cov(P00, C01, C11, c10=C10,backend=self.backend)

    def clean_norm(self):
        self.P1_dic = None
        self.P2_dic = None
        return

    def _compute_C01(self, j2, conv,
                     vmask, M_dic,
                     MconvPsi_dic):
        """
        Compute the C01 coefficients (auto or cross)
        C01 = < (Ia * Psi)_j3 x (|Ib * Psi_j2| * Psi_j3)^* >_pix
        Parameters
        ----------
        Returns
        -------
        cc01, sc01: real and imag parts of C01 coeff
        """
        ### Compute |I1 * Psi_j2| * Psi_j3 = M1_j2 * Psi_j3
        # Warning: M1_dic[j2] is already at j3 resolution [Nbatch, Npix_j3, Norient3]
        MconvPsi = self.convol(M_dic[j2], axis=1)  # [Nbatch, Npix_j3, Norient3, Norient2]

        # Store it so we can use it in C11 computation
        MconvPsi_dic[j2] = MconvPsi  # [Nbatch, Npix_j3, Norient3, Norient2]

        ### Compute the product (I2 * Psi)_j3 x (M1_j2 * Psi_j3)^*
        # z_1 x z_2^* = (a1a2 + b1b2) + i(b1a2 - a1b2)
        # cconv, sconv are [Nbatch, Npix_j3, Norient3]
        c01 = self.backend.bk_expand_dims(conv, -1) * self.backend.bk_conjugate(MconvPsi)   # [Nbatch, Npix_j3, Norient3, Norient2]

        ### Apply the mask [Nmask, Npix_j3] and sum over pixels
        c01 = self.masked_mean(c01, vmask, axis=1)  # [Nbatch, Nmask, Norient3, Norient2]
        return c01

    def _compute_C11(self, j1, j2, vmask,
                     M1convPsi_dic,
                     M2convPsi_dic=None):
        #### Simplify notations
        M1 = M1convPsi_dic[j1]  # [Nbatch, Npix_j3, Norient3, Norient1]
        
        # Auto or Cross coefficients
        if M2convPsi_dic is None:  # Auto
            M2 = M1convPsi_dic[j2]  # [Nbatch, Npix_j3, Norient3, Norient2]
        else:  # Cross
            M2 = M2convPsi_dic[j2]

        ### Compute the product (|I1 * Psi_j1| * Psi_j3)(|I2 * Psi_j2| * Psi_j3)
        # z_1 x z_2^* = (a1a2 + b1b2) + i(b1a2 - a1b2)
        c11 = self.backend.bk_expand_dims(M1, -2) * self.backend.bk_conjugate(self.backend.bk_expand_dims(M2, -1))  # [Nbatch, Npix_j3, Norient3, Norient2, Norient1]

        ### Apply the mask and sum over pixels
        c11 = self.masked_mean(c11, vmask, axis=1)  # [Nbatch, Nmask, Norient3, Norient2, Norient1]
        return c11

    def square(self, x):
        if isinstance(x, scat_cov):
            if x.S1 is None:
                return scat_cov(self.backend.bk_square(self.backend.bk_abs(x.P00)),
                                self.backend.bk_square(self.backend.bk_abs(x.C01)),
                                self.backend.bk_square(self.backend.bk_abs(x.C11)),backend=self.backend)
            else:
                return scat_cov(self.backend.bk_square(self.backend.bk_abs(x.P00)),
                                self.backend.bk_square(self.backend.bk_abs(x.C01)),
                                self.backend.bk_square(self.backend.bk_abs(x.C11)),
                                s1=self.backend.bk_square(self.backend.bk_abs(x.S1)),backend=self.backend)
        else:
            return self.backend.bk_abs(self.backend.bk_square(x))

    def sqrt(self, x):
        if isinstance(x, scat_cov):
            if x.S1 is None:
                return scat_cov(self.backend.bk_sqrt(self.backend.bk_abs(x.P00)),
                                self.backend.bk_sqrt(self.backend.bk_abs(x.C01)),
                                self.backend.bk_sqrt(self.backend.bk_abs(x.C11)),backend=self.backend)
            else:
                return scat_cov(self.backend.bk_sqrt(self.backend.bk_abs(x.P00)),
                                self.backend.bk_sqrt(self.backend.bk_abs(x.C01)),
                                self.backend.bk_sqrt(self.backend.bk_abs(x.C11)),
                                s1=self.backend.bk_sqrt(self.backend.bk_abs(x.S1)),backend=self.backend)
        else:
            return self.backend.bk_abs(self.backend.bk_sqrt(x))

    def reduce_mean(self, x):
        if isinstance(x, scat_cov):
            if x.S1 is None:
                result = (self.backend.bk_reduce_mean(x.P00) + \
                         self.backend.bk_reduce_mean(x.C01) + \
                         self.backend.bk_reduce_mean(x.C11))/3
            else:
                result = (self.backend.bk_reduce_mean(x.P00) + \
                         self.backend.bk_reduce_mean(x.S1) + \
                         self.backend.bk_reduce_mean(x.C01) + \
                         self.backend.bk_reduce_mean(x.C11))/4
        else:
            return self.backend.bk_reduce_mean(x)
        return result

    def reduce_sum(self, x):
        
        if isinstance(x, scat_cov):
            if x.S1 is None:
                result = self.backend.bk_reduce_sum(x.P00) + \
                         self.backend.bk_reduce_sum(x.C01) + \
                         self.backend.bk_reduce_sum(x.C11)
            else:
                result = self.backend.bk_reduce_sum(x.P00) + \
                         self.backend.bk_reduce_sum(x.S1) + \
                         self.backend.bk_reduce_sum(x.C01) + \
                         self.backend.bk_reduce_sum(x.C11)
        else:
            return self.backend.bk_reduce_sum(x)
        return result

    def log(self, x):
        if isinstance(x, scat_cov):

            if x.S1 is None:
                result = self.backend.bk_log(x.P00) + \
                         self.backend.bk_log(x.C01) + \
                         self.backend.bk_log(x.C11)
            else:
                result = self.backend.bk_log(x.P00) + \
                         self.backend.bk_log(x.S1) + \
                         self.backend.bk_log(x.C01) + \
                         self.backend.bk_log(x.C11)
        else:
            return self.backend.bk_log(x)
        
        return result
